"""
Facilitate user management operations such as changing owner,
Distinguished Name (DN) manipulation.

1. Connect to the environment using data from the workstation
2. Get first user group object based on provided name ('Second Factor Exempt')
3. Get second user group object based on provided name ('TransactionServer')
4. Move users from the first user group to the second user group
5. Change owner for all users to administrator
6. Get user group object based on provided name ('Second Factor Exempt')
7. Change/add LDAP Distinguished Name (DN) on target user group
"""

from mstrio.connection import get_connection
from mstrio.users_and_groups.user import User, list_users
from mstrio.users_and_groups.user_group import UserGroup

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# Move users from one user group to another
user_group_1 = UserGroup(conn, name='Second Factor Exempt')
user_group_2 = UserGroup(conn, name='TransactionServer')
user_group_2.add_users([user['id'] for user in user_group_1.members])
user_group_1.remove_all_users()

# change owner for all the users to Administrator
for user in list_users(conn):
    user.alter(owner=User(conn, name='Administrator'))

# change/add LDAP Distinguished Name (DN) on target user group
user_group = UserGroup(conn, name='Second Factor Exempt')
user_group.alter(ldapdn='CN=Users,DC=domain,DC=com')



import logging
from dataclasses import dataclass
from enum import auto
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import transformations
from mstrio.connection import Connection
from mstrio.modeling.expression import Expression, ExpressionFormat
from mstrio.modeling.schema import SchemaObjectReference
from mstrio.modeling.schema.helpers import ObjectSubType
from mstrio.object_management import search_operations
from mstrio.object_management.folder import Folder
from mstrio.object_management.search_enums import SearchPattern
from mstrio.types import ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    Dictable,
    delete_none_values,
    filter_params_for_func,
    find_object_with_name,
)
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0500')
def list_transformations(
    connection,
    name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    **filters,
) -> list["Transformation"] | list[dict]:
    """Get list of Transformation objects or dicts with them.

    Optionally use `to_dictionary` to choose output format.
    Optionally filter transformations by specifying 'name'.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name_begins = ?onny will return Sonny and Tonny

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
        name (string, optional): characters that the transformation name must
            begin with
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns Transformation object
        limit (integer, optional): limit the number of elements returned. If
            None all objects are returned.
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        search_pattern (SearchPattern enum or int, optional): pattern to
            search for, such as Begin With or Exactly. Possible values are
            available in ENUM `mstrio.object_management.SearchPattern`.
            Default value is CONTAINS (4).
        show_expression_as (ExpressionFormat, str): specify how expressions
            should be presented
            Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS or `tokens`
        **filters: Available filter parameters:
            id str: Transformation's ID
            name str: Transformation's name
            description str: Transformation's description
            date_created str: format: 2001-01-02T20:48:05.000+0000
            date_modified str: format: 2001-01-02T20:48:05.000+0000
            version str: Transformation's version
            owner dict: e.g. {'id': <user's id>, 'name': <user's name>},
                with one or both of the keys: id, name
            acg str | int: access control group
            subtype str: object's subtype
            ext_type str: object's extended type

    Returns:
        list with Transformation objects or list of dictionaries
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    objects_ = search_operations.full_search(
        connection,
        object_types=Transformation._OBJECT_TYPE,
        project=proj_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        **filters,
    )
    if to_dictionary:
        return objects_
    else:
        show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )
        return [
            Transformation.from_dict(
                source={**obj_, 'show_expression_as': show_expression_as},
                connection=connection,
                with_missing_value=True,
            )
            for obj_ in objects_
        ]


class MappingType(AutoName):
    """Enumeration constants used to specify mapping type"""

    ONE_TO_ONE = auto()
    MANY_TO_MANY = auto()


@class_version_handler('11.3.0500')
class Transformation(Entity, CopyMixin, MoveMixin, DeleteMixin):
    """Python representation of Strategy One Transformation object.

    Attributes:
        id: transformation's ID
        name: transformation's name
        sub_type: string literal used to identify the type of a metadata object
        description: transformation's description
        type: object type, ObjectTypes enum
        subtype: object subtype, ObjectSubTypes enum
        ext_type: object extended type, ExtendedType enum
        attributes: list of attributes used in the transformation
        date_created: creation time, DateTime object
        date_modified: last modification time, DateTime object
        mapping_type: transformation's mapping type, MappingType enum
        owner: User object that is the owner
        version_id: the version number this object is currently carrying
        primary_locale: The primary locale of the object, in the IETF BCP 47
            language tag format, such as "en-US".If no particular locale is set
            as the primary locale (for example, if the project is not
            internationalized), the field will be omitted.
        acg: access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: object access control list
        hidden: Specifies whether the object is hidden
    """

    _OBJECT_TYPE = ObjectTypes.ROLE
    _API_GETTERS = {
        (
            'id',
            'sub_type',
            'name',
            'description',
            'date_created',
            'date_modified',
            'primary_locale',
            'mapping_type',
            'attributes',
            'is_embedded',
        ): transformations.get_transformation,
        (
            'type',
            'subtype',
            'version',
            'owner',
            'acg',
            'acl',
            'ext_type',
            'hidden',
            'comments',
        ): objects_processors.get_info,
    }
    _API_PATCH = {
        (
            'name',
            'description',
            'mapping_type',
            'attributes',
            'destination_folder_id',
        ): (transformations.update_transformation, 'partial_put'),
        (
            'hidden',
            'comments',
            'owner',
        ): (objects_processors.update, "partial_put"),
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        "attributes": (
            lambda source, connection: [
                TransformationAttribute.from_dict(content, connection)
                for content in source
            ]
        ),
        "mapping_type": MappingType,
    }

    def __init__(
        self,
        connection: Connection,
        id=None,
        name=None,
        show_expression_as=ExpressionFormat.TREE,
    ):
        """Initializes a new instance of Transformation class

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            id (str, optional): Transformation's ID. Defaults to None.
            name (str, optional): Transformation's name. Defaults to None.
            show_expression_as (ExpressionFormat or str, optional):
                specify how expressions should be presented.
                Defaults to ExpressionFormat.TREE.

                Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS` or `tokens`

        Raises:
            AttributeError: if both `id` and `name` are not provided.
            ValueError: if Transformation with the given `name` doesn't exist.
        """
        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            transformation = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_transformations,
                search_pattern=SearchPattern.EXACTLY,
            )
            id = transformation['id']
        super().__init__(
            connection=connection,
            object_id=id,
            name=name,
            show_expression_as=show_expression_as,
        )

    def _init_variables(self, default_value, **kwargs) -> None:
        super()._init_variables(default_value=default_value, **kwargs)
        self._sub_type = kwargs.get('sub_type', default_value)
        self._is_embedded = kwargs.get('is_embedded', default_value)
        self._attributes = (
            [
                TransformationAttribute.from_dict(attr, self._connection)
                for attr in kwargs.get('attributes')
            ]
            if kwargs.get('attributes')
            else default_value
        )
        self._mapping_type = kwargs.get('mapping_type', default_value)
        self._version_id = kwargs.get('version_id')
        self._primary_locale = kwargs.get('primary_locale', default_value)
        show_expression_as = kwargs.get('show_expression_as', 'tree')
        self._show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )

    def list_properties(self, excluded_properties: list[str] | None = None) -> dict:
        excluded_properties = excluded_properties or []
        excluded_properties.extend(
            [
                'hidden',
                'icon_path',
                'comments',
                'certified_info',
                'project_id',
                'target_info',
                'view_media',
                'abbreviation',
            ]
        )
        return super().list_properties(excluded_properties=excluded_properties)

    @classmethod
    @method_version_handler('11.3.0500')
    def create(
        cls,
        connection: 'Connection',
        sub_type: ObjectSubType | str,
        name: str,
        destination_folder: Folder | str,
        attributes: list,
        mapping_type: MappingType | str,
        is_embedded: bool = False,
        description: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    ) -> 'Transformation':
        """Create Transformation object.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`
            sub_type: transformation's sub_type
            name: transformation's name
            destination_folder: A globally unique identifier used to
                distinguish between metadata objects within the same project.
                It is possible for two metadata objects in different projects
                to have the same Object Id.
            attributes: list of base transformation attributes
            mapping_type: transformation's mapping type
            is_embedded: If true indicates that the target object of this
                reference is embedded within this object. Alternatively if
                this object is itself embedded, then it means that the target
                object is embedded in the same container as this object.
            description: transformation's description
            show_expression_as (ExpressionFormat, str): specify how expressions
                should be presented
                Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS or `tokens`

        Returns:
            Transformation class object.
        """
        body = {
            'information': {
                'subType': get_enum_val(sub_type, ObjectSubType),
                'name': name,
                'isEmbedded': is_embedded,
                'description': description,
                'destinationFolderId': destination_folder,
            },
            'attributes': [attribute.to_dict() for attribute in attributes],
            'mappingType': mapping_type,
        }
        body = delete_none_values(body, recursion=True)
        response = transformations.create_transformation(
            connection=connection,
            body=body,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
        ).json()

        if config.verbose:
            logger.info(
                f"Successfully created transformation named: '{name}' with ID: '"
                f"{response['id']}'"
            )

        return cls.from_dict(
            source={**response, 'show_expression_as': show_expression_as},
            connection=connection,
        )

    def alter(
        self,
        name: str | None = None,
        destination_folder_id: Folder | str | None = None,
        attributes: list | None = None,
        mapping_type: MappingType | None = None,
        description: str | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ):
        """Alter transformation properties.

        Args:
            name: transformation's name
            destination_folder_id: A globally unique identifier used to
                distinguish between metadata objects within the same project.
                It is possible for two metadata objects in different projects
                to have the same Object Id.
            attributes: list of base transformation attributes
            mapping_type: transformation's mapping type
            description: transformation's description
            hidden (bool, optional): Specifies whether the object is hidden.
                Default value: False.
            comments: long description of the transformation
           owner: (str, User, optional): owner of the transformation
        """

        name = name or self.name
        if isinstance(owner, User):
            owner = owner.id
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)

    @property
    def sub_type(self):
        return self._sub_type

    @property
    def attributes(self):
        return self._attributes

    @property
    def mapping_type(self):
        return self._mapping_type

    @property
    def version_id(self):
        return self._version_id

    @property
    def primary_locale(self):
        return self._primary_locale

    @property
    def is_embedded(self):
        return self._is_embedded


@dataclass
class TransformationAttributeForm(Dictable):
    """Class representation of Transformation Attribute Form

    Attributes:
        id: A globally unique identifier used to distinguish between metadata
            objects within the same project. It is possible for two metadata
            objects in different projects to have the same Object Id.
        name: transformation attribute form name
        lookup_table: schema reference of a lookup table
        expression: A generic specification for a calculation stored within a
            metadata object. The expression is represented as a tree over nodes.
            Most internal nodes (called operator nodes) are defined by applying
            a function to the operator's child nodes.
    """

    _FROM_DICT_MAP = {"lookup_table": SchemaObjectReference, "expression": Expression}

    id: str
    name: str
    lookup_table: SchemaObjectReference
    expression: Expression

    def list_properties(self, camel_case=True) -> dict:
        """Lists all properties of transformation attribute form."""
        return self.to_dict(camel_case=camel_case)


@dataclass
class TransformationAttribute(Dictable):
    """Class representation of Transformation Attribute

    Attributes:
        id: A globally unique identifier used to distinguish between metadata
            objects within the same project. It is possible for two metadata
            objects in different projects to have the same Object Id.
        base_attribute: Object that specifies the base attribute of the
            transformation attribute.
        forms: Object that specifies the transformation attribute form.
    """

    _FROM_DICT_MAP = {
        "base_attribute": SchemaObjectReference,
        "forms": ([TransformationAttributeForm.from_dict]),
    }

    id: str
    base_attribute: SchemaObjectReference
    forms: list[TransformationAttributeForm]

    def list_properties(self, camel_case=True) -> dict:
        """Lists all properties of transformation attribute."""
        return self.to_dict(camel_case=camel_case)



import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

import pandas as pd
import pypika as sql

from mstrio import config
from mstrio.connection import Connection
from mstrio.datasources import DatasourceInstance
from mstrio.server.project import Project
from mstrio.types import ObjectTypes
from mstrio.utils.entity import Entity
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import (
    Dictable,
    IServerError,
    VersionException,
)
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors import objects, projects
from mstrio.utils.version_helper import is_server_min_version, method_version_handler

if TYPE_CHECKING:
    from mstrio.server.language import Language

logger = logging.getLogger(__name__)


@method_version_handler(version='11.3.7')
def list_translations(
    connection: Connection,
    id: str,
    object_type: int | ObjectTypes,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    languages: 'list[Language | str] | None' = None,
    to_dictionary: bool = False,
) -> list['Translation'] | list[dict]:
    """Lists translations of the given Object.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the Object the translations will be listed for
        object_type (int | ObjectTypes): type of the Object the translations
            will be listed for
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        languages (list, optional): list of languages to list the translations
            for, only translations from these languages will be listed.
            Languages in the list should be one of the following:
                - lcid attribute of the language
                - ID of the language
                - Language class object
        to_dictionary (bool, optional): If True returns a list of dictionaries,
            by default (False) returns a list of Translation objects

    Returns:
        A list of dictionaries representing translations for the Object or a
        list of Translation Objects.
    """
    object_type = get_enum_val(object_type, ObjectTypes)

    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
        assert_id_exists=False,
    )

    output = objects.get_translations(
        connection=connection,
        project_id=proj_id,
        target_id=id,
        object_type=object_type,
    )

    if languages:
        proper_language_list = [
            Translation._get_lang_lcid(connection=connection, language=lang)
            for lang in languages
        ]
        new_output = []

        for target_id in output:
            values = target_id.get('translation_values')
            new_values = [
                value
                for value in values
                if int(value.get('language_lcid')) in proper_language_list
            ]

            new_output.append(
                {
                    'translation_target_id': target_id.get('translation_target_id'),
                    'translation_target_name': target_id.get('translation_target_name'),
                    'translation_values': new_values,
                }
            )

        output = new_output

    if to_dictionary:
        return output
    return [
        Translation.from_dict(source=translation_data, to_snake_case=False)
        for translation_data in output
    ]


@dataclass
class Translation(Dictable):
    """Translation class represents a translation for an Object. It also
    contains methods to create, alter and delete translations for all Object
    types including ones not yet supported by mstrio-py

    Attributes:
            translation_target_id (str): ID of the part of the Object the
                translations are for
            translation_target_name (str): name of the part of the Object the
                translations are for
            translation_values (list[Translation.TranslationValues]): list of
                translation values for the part of the Object"""

    _db_types = {
        'MySQL': sql.MySQLQuery,
        'MSSQL': sql.MSSQLQuery,
        'PostgreSQL': sql.PostgreSQLQuery,
        'Oracle': sql.OracleQuery,
        'Vertica': sql.VerticaQuery,
    }

    @dataclass
    class TranslationValue(Dictable):
        """Object that specifies a Translation for a single Language

        Attributes:
            value (str): value of the translation
            language_lcid (str): LCID of the language
        """

        value: str
        language_lcid: str

    @dataclass
    class OperationData:
        """Object that specifies data for a single Operation on a Translation.

        Attributes:
            target_language (Language | str | int): one of the following:
                - lcid attribute of the language
                - ID of the language
                - Language class object
            target_id (str): ID of the specific part of the Object to be
                translated
            value (str, optional): value of the change, not necessary when
                deleting a translation
        """

        from mstrio.server.language import Language

        target_language: Language | str | int
        target_id: str
        value: str | None = None

    _FROM_DICT_MAP = {
        'translation_values': lambda source, connection: [
            Translation.TranslationValue.from_dict(content, connection)
            for content in source
        ],
    }

    translation_target_id: str
    translation_target_name: str
    translation_values: list[TranslationValue]

    @staticmethod
    @method_version_handler(version='11.3.7')
    def add_translation(
        connection: Connection,
        id: str,
        object_type: int | ObjectTypes,
        translations: list[OperationData],
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> list['Translation']:
        """Adds translations to the Object.

        Args:
            connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
            id (str): ID of the Object the translation will be added to
            object_type (int | ObjectTypes): type of the Object
            translations (list[OperationData]): list of translations to be added
                to the Object
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name

        Returns:
            A list of translations for the Object.
        """
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        object_type = get_enum_val(object_type, ObjectTypes)
        operations_list = []
        translations = Translation._get_valid_target_ids(
            connection=connection,
            translations=translations,
            id=id,
            object_type=object_type,
            project_id=proj_id,
            operation_type='adding',
        )

        for translation in translations:
            base_lcid = Translation._get_lang_lcid(
                connection=connection, language=translation.target_language
            )
            path = f"/localesAndTranslations/{translation.target_id}/translationValues"
            operation = {
                'op': 'add',
                'path': path,
                'value': {base_lcid: {'translation': translation.value}},
            }
            operations_list.append(operation)

        body = {'operationList': operations_list}

        output = objects.update_translations(
            connection=connection,
            target_id=id,
            project_id=proj_id,
            body=body,
            object_type=object_type,
        )
        if config.verbose:
            logger.info(f"Successfully added Translations to Object with ID: {id}")
        return [
            Translation.from_dict(source=translation_data, to_snake_case=False)
            for translation_data in output
        ]

    @staticmethod
    @method_version_handler(version='11.3.7')
    def alter_translation(
        connection: Connection,
        id: str,
        object_type: int | ObjectTypes,
        translations: list[OperationData],
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> None:
        """Alters translations for the Object.

        Args:
            connection(Connection): Strategy One connection object returned by
            `connection.Connection()`
            id (str): ID of the Object the translation will be added to
            object_type (int | ObjectTypes): type of the Object
            translations (list[OperationData]): list of translations to be added
                to the Object
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        object_type = get_enum_val(object_type, ObjectTypes)
        operations_list = []
        obj_translations = list_translations(
            connection=connection,
            id=id,
            object_type=object_type,
            project_id=proj_id,
        )
        translations = Translation._get_valid_target_ids(
            connection=connection,
            translations=translations,
            id=id,
            object_type=object_type,
            project_id=proj_id,
            operation_type='altering',
        )

        for translation in translations:
            base_lcid = Translation._get_lang_lcid(
                connection=connection, language=translation.target_language
            )

            for translation_check in obj_translations:
                if translation_check.translation_target_id == translation.target_id:
                    existing_translation_languages = [
                        value.language_lcid
                        for value in translation_check.translation_values
                    ]
                    if str(base_lcid) not in existing_translation_languages:
                        error_msg = (
                            "Cannot alter translation for Language with LCID: "
                            f"{base_lcid} as there is no translation present in "
                            f"the target_id: {translation.target_id} for the language."
                        )
                        raise ValueError(error_msg)
                    else:
                        break

            path = (
                f"/localesAndTranslations/{translation.target_id}"
                f"/translationValues/{base_lcid}/translation"
            )
            operation = {
                'op': 'replace',
                'path': path,
                'value': translation.value,
            }
            operations_list.append(operation)

        body = {'operationList': operations_list}

        objects.update_translations(
            connection=connection,
            target_id=id,
            project_id=proj_id,
            body=body,
            object_type=object_type,
        )

        if config.verbose:
            logger.info(f"Successfully altered Translations of Object with ID: {id}")

    @staticmethod
    @method_version_handler(version='11.3.7')
    def remove_translation(
        connection: Connection,
        id: str,
        object_type: int | ObjectTypes,
        translations: list[OperationData],
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> None:
        """Removes translations from the Object.

        Args:
            connection(Connection): Strategy One connection object returned by
            `connection.Connection()`
            id (str): ID of the Object the translation will be added to
            object_type (int | ObjectTypes): type of the Object
            translations (list[OperationData]): list of translations to be added
                to the Object
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        object_type = get_enum_val(object_type, ObjectTypes)
        operations_list = []
        translations = Translation._get_valid_target_ids(
            connection=connection,
            translations=translations,
            id=id,
            object_type=object_type,
            project_id=proj_id,
            operation_type='deleting',
        )

        for translation in translations:
            base_lcid = Translation._get_lang_lcid(
                connection=connection, language=translation.target_language
            )
            path = (
                f"/localesAndTranslations/{translation.target_id}"
                f"/translationValues/{base_lcid}"
            )
            operation = {
                'op': 'remove',
                'path': path,
            }
            operations_list.append(operation)

        body = {'operationList': operations_list}

        objects.update_translations(
            connection=connection,
            target_id=id,
            project_id=proj_id,
            body=body,
            object_type=object_type,
        )

        if config.verbose:
            logger.info(f"Successfully deleted Translations for Object with ID: {id}")

    @staticmethod
    @method_version_handler(version='11.3.7')
    def to_csv_from_list(
        connection: Connection,
        object_list: list[Entity],
        separator: str = ';',
        languages: list | None = None,
        file_path: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        denormalized_form: bool = False,
        add_object_path: bool = False,
        add_object_description: bool = False,
        add_object_version: bool = False,
        add_object_last_modified_date: bool = False,
        add_object_creation_date: bool = False,
    ) -> str | None:
        """Export translations of the given objects to a CSV file.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            object_list (list[Entity]): list of Objects to export translations
                for. Objects have to be provided as class objects such as a
                Metric or a Document.
            separator (str, optional): specify the separator for the csv file.
                Defaults to a semicolon ;
            languages (list, optional): list of languages to list translations
                for, only translations from these languages will be listed.
                If not specified languages will be taken from the project.
                Languages in the list should be one of the following:
                    - ID of the language
                    - Language class object
            file_path (str, optional): a path specifying where to save the csv
                file
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            denormalized_form (bool, optional): if True exports the data in a
                denormalized form, if False exports the data in a normalized
                form. False by default.
                Normalised form means that each row contains a single target_id
                with translation values for all required languages
                Denormalised form means each row represents a translation value
                for one target_id and one language at a time
            add_object_path (bool, optional): if True adds a column with the
                path of the Object
            add_object_description (bool, optional): if True adds a column with
                the description of the Object
            add_object_version (bool, optional): if True adds a column with the
                version of the Object
            add_object_last_modified_date (bool, optional): if True adds a
                column with the last modified date of the Object
            add_object_creation_date (bool, optional): if True adds a column
                with the creation date of the Object

        Returns:
            String representation of the CSV file if no path was given or
            `None` if path was provided.
        """
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        _prepare_dataframe = (
            Translation._prepare_denormalized_dataframe
            if denormalized_form
            else Translation._prepare_normalized_dataframe
        )
        dataframe = _prepare_dataframe(
            connection=connection,
            project_id=proj_id,
            object_list=object_list,
            languages=languages,
            path=add_object_path,
            description=add_object_description,
            version=add_object_version,
            last_modified=add_object_last_modified_date,
            creation_date=add_object_creation_date,
        )
        return dataframe.to_csv(index=False, path_or_buf=file_path, sep=separator)

    @staticmethod
    @method_version_handler(version='11.3.7')
    def add_translations_from_csv(
        connection: Connection,
        file_path: str,
        separator: str = ';',
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        delete: bool = False,
        automatch_target_ids: bool = False,
    ) -> None:
        """Add translations from a specified CSV file.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            file_path (str): a path specifying the CSV file
            separator (str, optional): specify the separator of the csv file.
                Defaults to a semicolon ;
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            delete (bool, optional): if True deletes translations for languages
                that are empty in the CSV file. False by default.
            automatch_target_ids (bool, optional): if True tries to match the
                target IDs from the file that are not found in the Objects on
                the server by the content of the translation for the project's
                default language. False by default.

        To ensure correct import, the columns from the original file cannot be
        renamed. Columns can be deleted (with the exception of object ID, object
        type and target ID) but new columns cannot be added. The column order
        does not matter.
        """

        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )
        df = pd.read_csv(file_path, na_filter=False, sep=separator)

        Translation._implement_changes_from_dataframe(
            connection=connection,
            df=df,
            project_id=proj_id,
            delete=delete,
            automatch_target_ids=automatch_target_ids,
        )
        if config.verbose:
            logger.info("Successfully implemented changes from the CSV file.")

    @staticmethod
    @method_version_handler(version='11.3.7')
    def to_database_from_list(
        connection: Connection,
        object_list: list[Entity],
        table_name: str,
        datasource: DatasourceInstance | str,
        database_type: str | None = None,
        languages: list | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        denormalized_form: bool = False,
        add_object_path: bool = False,
        add_object_description: bool = False,
        add_object_version: bool = False,
        add_object_last_modified_date: bool = False,
        add_object_creation_date: bool = False,
        force: bool = False,
    ) -> None:
        """Export translations of the given objects to an SQL database table.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            object_list (list[Entity]): list of Objects to export translations
                for. Objects have to be provided as class objects such as a
                Metric or a Document.
            table_name (str): name of the table to export the data to
                If the table with the given name already exists, the existing
                table will be dropped first, then a new one will be created.
            datasource (DatasourceInstance | str): DatasourceInstance object or
                ID of the DatasourceInstance containing the connection to the
                database to export the data to
            database_type (str, optional): type of the database, if not provided
                standard SQL will be used.
                Supported database types include:
                    - MySQL
                    - MSSQL
                    - PostgreSQL
                    - Oracle
                    - Vertica
            languages (list, optional): list of languages to list translations
                for, only translations from these languages will be listed.
                If not specified languages will be taken from the project.
                Languages in the list should be one of the following:
                    - ID of the language
                    - Language class object
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            denormalized_form (bool, optional): if True exports the data in a
                denormalized form, if False exports the data in a normalized
                form. False by default.
                Normalised form means that each row contains a single target_id
                with translation values for all required languages
                Denormalised form means each row represents a translation value
                for one target_id and one language at a time
            add_object_path (bool, optional): if True adds a column with the
                path of the Object
            add_object_description (bool, optional): if True adds a column with
                the description of the Object
            add_object_version (bool, optional): if True adds a column with the
                version of the Object
            add_object_last_modified_date (bool, optional): if True adds a
                column with the last modified date of the Object
            add_object_creation_date (bool, optional): if True adds a column
                with the creation date of the Object
            force (bool, optional): if True skips the prompt asking for
                confirmation before dropping the table. False by default.
        """
        if not force:
            user_input = input(
                "Warning: This method will drop the specified table if "
                "it exists before exporting the translations, proceed? [Y/N]: "
            )
            if user_input != "Y":
                return None

        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        if isinstance(datasource, DatasourceInstance):
            datasource = datasource.id
        _prepare_dataframe = (
            Translation._prepare_denormalized_dataframe
            if denormalized_form
            else Translation._prepare_normalized_dataframe
        )
        dataframe = _prepare_dataframe(
            connection=connection,
            project_id=proj_id,
            object_list=object_list,
            languages=languages,
            path=add_object_path,
            description=add_object_description,
            version=add_object_version,
            last_modified=add_object_last_modified_date,
            creation_date=add_object_creation_date,
        )
        if config.verbose:
            logger.info("Getting query status on dropping table.")
        DatasourceInstance._execute_query(
            connection=connection,
            query=sql.Query.drop_table(table_name).if_exists(),
            datasource_id=datasource,
            project_id=proj_id,
        )

        if config.verbose:
            logger.info("Getting query status on creating table.")
        DatasourceInstance._execute_query(
            connection=connection,
            query=Translation._create_table_query(
                dataframe=dataframe,
                table_name=table_name,
                database_type=database_type,
            ),
            datasource_id=datasource,
            project_id=proj_id,
        )

        if config.verbose:
            logger.info("Getting query status on populating table.")
        DatasourceInstance._execute_query(
            connection=connection,
            query=Translation._create_data_query(
                dataframe=dataframe,
                table_name=table_name,
                database_type=database_type,
            ),
            datasource_id=datasource,
            project_id=proj_id,
        )

        if config.verbose:
            logger.info("Successfully exported translations to the database.")

    @staticmethod
    @method_version_handler(version='11.3.7')
    def add_translations_from_database(
        connection: Connection,
        table_name: str,
        datasource: DatasourceInstance | str,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        delete: bool = False,
        automatch_target_ids: bool = False,
    ) -> None:
        """Add translations from a specified database table.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            table_name (str): name of the table to import the data from
            datasource (DatasourceInstance | str): DatasourceInstance object or
                ID of the DatasourceInstance containing the connection to the
                database to export the data to
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            delete (bool, optional): if True deletes translations for languages
                that are empty in the database table. False by default.
            automatch_target_ids (bool, optional): if True tries to match the
                target IDs from the file that are not found in the Objects on
                the server by the content of the translation for the project's
                default language. False by default.

        To ensure correct import, the columns originally exported to database
        cannot be renamed. Columns can be deleted (with the exception of object
        ID, object type and target ID) but new columns cannot be added.
        """

        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        if isinstance(datasource, DatasourceInstance):
            datasource = datasource.id
        if config.verbose:
            logger.info("Fetching translation data from the table.")
        execution_data = DatasourceInstance._execute_query(
            connection=connection,
            query=sql.Query.from_(table_name).select('*'),
            datasource_id=datasource,
            project_id=proj_id,
        )
        df = pd.DataFrame.from_dict(
            data=execution_data.get('results').get('data'), orient='columns'
        )
        Translation._implement_changes_from_dataframe(
            connection=connection,
            df=df,
            project_id=proj_id,
            delete=delete,
            automatch_target_ids=automatch_target_ids,
        )
        if config.verbose:
            logger.info("Successfully implemented changes from the database.")

    @staticmethod
    @method_version_handler(version='11.3.7')
    def to_json_from_list(
        connection: Connection,
        object_list: list[Entity],
        languages: list | None = None,
        file_path: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        denormalized_form: bool = False,
        add_object_path: bool = False,
        add_object_description: bool = False,
        add_object_version: bool = False,
        add_object_last_modified_date: bool = False,
        add_object_creation_date: bool = False,
    ) -> str | None:
        """Export translations of the given objects to a JSON file.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            object_list (list[Entity]): list of Objects to export translations
                for. Objects have to be provided as class objects such as a
                Metric or a Document.
            languages (list, optional): list of languages to list translations
                for, only translations from these languages will be listed.
                If not specified languages will be taken from the project.
                Languages in the list should be one of the following:
                    - ID of the language
                    - Language class object
            file_path (str, optional): a path specifying where to save the
                json file
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            denormalized_form (bool, optional): if True exports the data in a
                denormalized form, if False exports the data in a normalized
                form. False by default.
                Normalised form means that each row contains a single target_id
                with translation values for all required languages
                Denormalised form means each row represents a translation value
                for one target_id and one language at a time
            add_object_path (bool, optional): if True adds a column with the
                path of the Object
            add_object_description (bool, optional): if True adds a column with
                the description of the Object
            add_object_version (bool, optional): if True adds a column with the
                version of the Object
            add_object_last_modified_date (bool, optional): if True adds a
                column with the last modified date of the Object
            add_object_creation_date (bool, optional): if True adds a column
                with the creation date of the Object

        Returns:
            String representation of the JSON file if no path was given or
            `None` if path was provided.
        """

        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        _prepare_dataframe = (
            Translation._prepare_denormalized_dataframe
            if denormalized_form
            else Translation._prepare_normalized_dataframe
        )
        dataframe = _prepare_dataframe(
            connection=connection,
            project_id=proj_id,
            object_list=object_list,
            languages=languages,
            path=add_object_path,
            description=add_object_description,
            version=add_object_version,
            last_modified=add_object_last_modified_date,
            creation_date=add_object_creation_date,
        )
        if file_path:
            with open(file=file_path, mode='w', encoding='utf-8') as file:
                return dataframe.to_json(
                    path_or_buf=file,
                    orient='records',
                    force_ascii=False,
                    date_format='iso',
                )
        else:
            return dataframe.to_json(orient='records', date_format='iso')

    @staticmethod
    @method_version_handler(version='11.3.7')
    def add_translations_from_json(
        connection: Connection,
        file_path: str,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        delete: bool = False,
        automatch_target_ids: bool = False,
    ) -> None:
        """Add translations from a specified JSON file.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            file_path (str): a path specifying the JSON file
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            delete (bool, optional): if True deletes translations for languages
                that are empty in the JSON file. False by default.
            automatch_target_ids (bool, optional): if True tries to match the
                target IDs from the file that are not found in the Objects on
                the server by the content of the translation for the project's
                default language. False by default.

        To ensure correct import, the columns from the original file cannot be
        renamed. Columns can be deleted (with the exception of object ID, object
        type and target ID) but new columns cannot be added. The column order
        does not matter.
        """
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )
        df = pd.read_json(file_path, orient='records')

        Translation._implement_changes_from_dataframe(
            connection=connection,
            df=df,
            project_id=proj_id,
            delete=delete,
            automatch_target_ids=automatch_target_ids,
        )

        if config.verbose:
            logger.info("Successfully implemented changes from the JSON file.")

    @staticmethod
    @method_version_handler(version='11.3.7')
    def to_dataframe_from_list(
        connection: Connection,
        object_list: list[Entity],
        languages: list | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        denormalized_form: bool = False,
        add_object_path: bool = False,
        add_object_description: bool = False,
        add_object_version: bool = False,
        add_object_last_modified_date: bool = False,
        add_object_creation_date: bool = False,
    ) -> pd.DataFrame:
        """Export translations of the given objects to a dataframe.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            object_list (list[Entity]): list of Objects to export translations
                for. Objects have to be provided as class objects such as a
                Metric or a Document.
            languages (list, optional): list of languages to list translations
                for, only translations from these languages will be listed.
                If not specified languages will be taken from the project.
                Languages in the list should be one of the following:
                    - ID of the language
                    - Language class object
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            denormalized_form (bool, optional): if True exports the data in a
                denormalized form, if False exports the data in a normalized
                form. False by default.
                Normalised form means that each row contains a single target_id
                with translation values for all required languages
                Denormalised form means each row represents a translation value
                for one target_id and one language at a time
            add_object_path (bool, optional): if True adds a column with the
                path of the Object
            add_object_description (bool, optional): if True adds a column with
                the description of the Object
            add_object_version (bool, optional): if True adds a column with the
                version of the Object
            add_object_last_modified_date (bool, optional): if True adds a
                column with the last modified date of the Object
            add_object_creation_date (bool, optional): if True adds a column
                with the creation date of the Object

        Returns:
            A pandas Dataframe containing the translation data.
        """
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )
        _prepare_dataframe = (
            Translation._prepare_denormalized_dataframe
            if denormalized_form
            else Translation._prepare_normalized_dataframe
        )
        dataframe = _prepare_dataframe(
            connection=connection,
            project_id=proj_id,
            object_list=object_list,
            languages=languages,
            path=add_object_path,
            description=add_object_description,
            version=add_object_version,
            last_modified=add_object_last_modified_date,
            creation_date=add_object_creation_date,
        )
        return dataframe

    @staticmethod
    @method_version_handler(version='11.3.7')
    def add_translations_from_dataframe(
        connection: Connection,
        dataframe: pd.DataFrame,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        delete: bool = False,
        automatch_target_ids: bool = False,
    ) -> None:
        """Add translations from a given dataframe.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            dataframe (pd.DataFrame): dataframe containing the data to be
                imported
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            delete (bool, optional): if True deletes translations for languages
                that are empty in the dataframe. False by default.
            automatch_target_ids (bool, optional): if True tries to match the
                target IDs from the file that are not found in the Objects on
                the server by the content of the translation for the project's
                default language. False by default.

        To ensure correct import, the columns from the original dataframe cannot
        be renamed. Columns can be deleted (with the exception of object ID,
        object type and target ID) but new columns cannot be added. The column
        order does not matter.
        """
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        Translation._implement_changes_from_dataframe(
            connection=connection,
            df=dataframe,
            project_id=proj_id,
            delete=delete,
            automatch_target_ids=automatch_target_ids,
        )

        if config.verbose:
            logger.info("Successfully implemented changes from the dataframe.")

    @staticmethod
    def _get_lang_lcid(connection: Connection, language) -> int:
        """Returns the lcid of the language.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            language (str | int | Language): language to get the lcid for, can
                be one of the following:
                    - lcid attribute of the language (will be returned back)
                    - ID of the language
                    - Language class object

        Returns:
            lcid of the language.
        """
        from mstrio.server.language import Language

        if isinstance(language, int):
            return language
        elif isinstance(language, str):
            return Language(connection=connection, id=language).lcid
        elif isinstance(language, Language):
            return language.lcid
        else:
            raise ValueError("Please provide a valid base language.")

    @staticmethod
    def _create_table_query(
        dataframe: pd.DataFrame,
        table_name: str,
        database_type: str | None = None,
    ) -> sql.Query:
        """Creates a query for creating a table in the database.

        Args:
            dataframe (pd.DataFrame): dataframe containing the data on the
                columns necessary for the table
            table_name (str): name of the table to be created
            database_type (str, optional): type of the database, if not provided
                standard SQL will be used.

        Returns:
            Prepared pypika query.
        """
        columns = []
        for column in dataframe.columns.values:
            columns.append(sql.Column(column, 'VARCHAR'))
        if database_type is None:
            return sql.Query.create_table(table_name).columns(*columns)
        else:
            return (
                Translation._db_types.get(database_type)
                .create_table(table_name)
                .columns(*columns)
            )

    @staticmethod
    def _create_data_query(
        dataframe: pd.DataFrame,
        table_name: str,
        database_type: str | None = None,
    ) -> sql.Query:
        """Creates a query for inserting data into the database.

        Args:
            dataframe (pd.DataFrame): dataframe containing the data to be
                inserted into the table
            table_name (str): name of the table to insert the data into
            database_type (str, optional): type of the database, if not provided
                standard SQL will be used.

        Returns:
            Prepared pypika Query.
        """
        tuples = list(dataframe.itertuples(index=False, name=None))
        if database_type is None:
            return sql.Query.into(table_name).insert(*tuples)
        else:
            return (
                Translation._db_types.get(database_type)
                .into(table_name)
                .insert(*tuples)
            )

    @staticmethod
    def _prepare_normalized_dataframe(
        connection: Connection,
        project_id: str,
        object_list: list[Entity],
        languages: list | None = None,
        path: bool = False,
        description: bool = False,
        version: bool = False,
        last_modified: bool = False,
        creation_date: bool = False,
    ) -> pd.DataFrame:
        """Prepares a normalized dataframe for exporting translations.
        Normalised form means that each row contains a single target_id with
        translation values for all required languages.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            project_id (str): ID of the project the Objects are a part of
            object_list (list[Entity]): list of Objects to export translations
                for. Objects have to be provided as class objects such as a
                Metric or a Document.
            languages (list, optional): list of languages to list translations
                for, only translations from these languages will be listed.
                If not specified languages will be taken from the project.
                Languages in the list should be one of the following:
                    - ID of the language
                    - Language class object
            path (bool, optional): if True adds a column with the path of the
                Object
            description (bool, optional): if True adds a column with the
                description of the Object
            version (bool, optional): if True adds a column with the version of
                the Object
            last_modified (bool, optional): if True adds a column with the last
                modified date of the Object
            creation_date (bool, optional): if True adds a column with the
                creation date of the Object

        Returns:
            Dataframe containing the data to be exported.
        """
        from mstrio.server.language import Language

        if not object_list:
            raise ValueError(
                'List of objects to export translations for is empty. '
                'Please make sure to provide at least one object in object_list.'
            )

        languages_list = []
        languages_list_lcid = []
        dataframes = []
        project_name = Project(connection, id=project_id).name

        if languages:
            for lang in languages:
                if isinstance(lang, str):
                    temp_lang = Language(connection=connection, id=lang)
                    languages_list.append(temp_lang.name)
                    languages_list_lcid.append(temp_lang.lcid)
                elif isinstance(lang, Language):
                    languages_list.append(lang.name)
                    languages_list_lcid.append(lang.lcid)
        else:
            proj_languages = projects.get_project_languages(
                connection=connection, id=project_id, path='metadata'
            )
            for lang in proj_languages:
                languages_list.append(proj_languages.get(lang).get('name'))
                languages_list_lcid.append(
                    Translation._get_lang_lcid(connection=connection, language=lang)
                )
        for curr_object in object_list:
            object_type = get_enum_val(curr_object.type, ObjectTypes)

            translations = list_translations(
                connection=connection,
                id=curr_object.id,
                object_type=object_type,
                project_id=project_id,
                languages=languages_list_lcid,
            )
            columns_list = [
                'project ID',
                'project name',
                'object ID',
                'object name',
                'object type',
            ]
            if path:
                columns_list.append('object path')
            if description:
                columns_list.append('object description')
            if version:
                columns_list.append('object version')
            if last_modified:
                columns_list.append('last modified date')
            if creation_date:
                columns_list.append('creation date')
            columns_list += ['target name', 'target ID']
            for translation in translations:
                row_to_add = [
                    project_id,
                    project_name,
                    curr_object.id,
                    curr_object.name,
                    object_type,
                ]
                if path:
                    row_to_add.append(curr_object.location)
                if description:
                    row_to_add.append(curr_object.description)
                if version:
                    row_to_add.append(curr_object.version)
                if last_modified:
                    row_to_add.append(curr_object.date_modified)
                if creation_date:
                    row_to_add.append(curr_object.date_created)
                row_to_add += [
                    translation.translation_target_name,
                    translation.translation_target_id,
                ]
                for lang in languages_list_lcid:
                    value_to_put = ''
                    for trans_value in translation.translation_values:
                        if int(trans_value.language_lcid) == lang:
                            value_to_put = trans_value.value
                            break
                    row_to_add.append(value_to_put)
                dataframes.append(
                    pd.DataFrame(
                        data=[row_to_add],
                        columns=columns_list + languages_list,
                    ),
                )
        dataframe = pd.concat(dataframes, ignore_index=True)
        return dataframe

    @staticmethod
    def _prepare_denormalized_dataframe(
        connection: Connection,
        project_id: str,
        object_list: list[Entity],
        languages: list | None = None,
        path: bool = False,
        description: bool = False,
        version: bool = False,
        last_modified: bool = False,
        creation_date: bool = False,
    ) -> pd.DataFrame:
        """Prepares a denormalized dataframe for exporting translations.
        Denormalised form means each row represents a translation value for one
        target_id and one language at a time.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            project_id (str): ID of the project the Objects are a part of
            object_list (list[Entity]): list of Objects to export translations
                for. Objects have to be provided as class objects such as a
                Metric or a Document.
            languages (list, optional): list of languages to list translations
                for, only translations from these languages will be listed.
                If not specified languages will be taken from the project.
                Languages in the list should be one of the following:
                    - ID of the language
                    - Language class object
            path (bool, optional): if True adds a column with the path of the
                Object
            description (bool, optional): if True adds a column with the
                description of the Object
            version (bool, optional): if True adds a column with the version of
                the Object
            last_modified (bool, optional): if True adds a column with the last
                modified date of the Object
            creation_date (bool, optional): if True adds a column with the
                creation date of the Object

        Returns:
            Dataframe containing the data to be exported.
        """
        from mstrio.server.language import Language

        if not object_list:
            raise ValueError(
                'List of objects to export translations for is empty. '
                'Please make sure to provide at least one object in object_list.'
            )

        languages_list = []
        languages_list_lcid = []
        dataframes = []
        project_name = Project(connection, id=project_id).name

        if languages:
            for lang in languages:
                if isinstance(lang, str):
                    temp_lang = Language(connection=connection, id=lang)
                    languages_list.append(temp_lang.name)
                    languages_list_lcid.append(temp_lang.lcid)
                elif isinstance(lang, Language):
                    languages_list.append(lang.name)
                    languages_list_lcid.append(lang.lcid)
        else:
            proj_languages = projects.get_project_languages(
                connection=connection, id=project_id, path='metadata'
            )
            for lang in proj_languages:
                languages_list.append(proj_languages.get(lang).get('name'))
                languages_list_lcid.append(
                    Translation._get_lang_lcid(connection=connection, language=lang)
                )

        for curr_object in object_list:
            object_type = get_enum_val(curr_object.type, ObjectTypes)

            translations = list_translations(
                connection=connection,
                id=curr_object.id,
                object_type=object_type,
                project_id=project_id,
                languages=languages_list_lcid,
            )
            columns_list = [
                'project ID',
                'project name',
                'object ID',
                'object name',
                'object type',
            ]
            if path:
                columns_list.append('object path')
            if description:
                columns_list.append('object description')
            if version:
                columns_list.append('object version')
            if last_modified:
                columns_list.append('last modified date')
            if creation_date:
                columns_list.append('creation date')
            columns_list += [
                'target name',
                'target ID',
                'language name',
                'translation value',
            ]
            for translation in translations:
                for trans_value in translation.translation_values:
                    if int(trans_value.language_lcid) in languages_list_lcid:
                        row_to_add = [
                            project_id,
                            project_name,
                            curr_object.id,
                            curr_object.name,
                            object_type,
                        ]
                        if path:
                            row_to_add.append(curr_object.location)
                        if description:
                            row_to_add.append(curr_object.description)
                        if version:
                            row_to_add.append(curr_object.version)
                        if last_modified:
                            row_to_add.append(curr_object.date_modified)
                        if creation_date:
                            row_to_add.append(curr_object.date_created)
                        row_to_add += [
                            translation.translation_target_name,
                            translation.translation_target_id,
                            languages_list[
                                languages_list_lcid.index(
                                    int(trans_value.language_lcid)
                                )
                            ],
                            trans_value.value,
                        ]
                        dataframes.append(
                            pd.DataFrame(
                                data=[row_to_add],
                                columns=columns_list,
                            ),
                        )
        dataframe = pd.concat(dataframes, ignore_index=True)
        return dataframe

    @staticmethod
    def _implement_changes_from_dataframe(
        connection: Connection,
        df: pd.DataFrame,
        project_id: str,
        delete: bool = False,
        automatch_target_ids: bool = False,
    ) -> None:
        """Implements changes to translations from a dataframe.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            df (pd.DataFrame): dataframe containing the changes to be applied
                to the translations on the server
            project_id (str): ID of the project the Objects are a part of
            delete (bool, optional): if True deletes translations for languages
                that are empty in the dataframe. False by default.
            automatch_target_ids (bool, optional): if True tries to match the
                target IDs from the file that are not found in the Objects on
                the server by the content of the translation for the project's
                default language. False by default.
        """
        from mstrio.server.language import Language

        if delete:
            if not is_server_min_version(
                connection=connection, version_str='11.4.0300'
            ):
                raise VersionException(
                    'Support for deletion requires version 11.4.03 or later'
                )
            operations_delete = []

        if 'language name' in df.columns.tolist():
            raise KeyError('Denormalized form of the dataframe is not supported.')

        if automatch_target_ids:
            project_language = Language(
                connection=connection,
                id=list(
                    projects.get_default_language(
                        connection=connection, id=project_id, path='metadata'
                    ).keys()
                )[0],
            ).lcid

        objects = []
        object_types = []
        operations = []

        intended_columns = [
            'project ID',
            'project name',
            'object ID',
            'object name',
            'object type',
            'object path',
            'object description',
            'object version',
            'last modified date',
            'creation date',
            'target name',
            'target ID',
        ]
        # Code below ensures that the columns are in the correct order specified
        # in the intended_columns list, followed by the languages columns
        columns = [
            column for column in intended_columns if column in df.columns.tolist()
        ]
        df = df[columns + df.columns.drop(columns).tolist()]

        columns = df.columns.tolist()
        languages = [
            Language(connection=connection, name=lang).lcid
            for lang in columns[columns.index('target ID') + 1 :]
        ]
        rows = df.values.tolist()
        for row in rows:
            if row[columns.index('object ID')] not in objects:
                objects.append(row[columns.index('object ID')])
                object_types.append(int(row[columns.index('object type')]))
                operations.append([])
                if delete:
                    operations_delete.append([])
            for index, value in enumerate(row[columns.index('target ID') + 1 :]):
                if value != '':
                    operations[objects.index(row[columns.index('object ID')])].append(
                        Translation.OperationData(
                            target_language=languages[index],
                            target_id=row[columns.index('target ID')],
                            value=value,
                        )
                    )
                elif delete:
                    operations_delete[
                        objects.index(row[columns.index('object ID')])
                    ].append(
                        Translation.OperationData(
                            target_language=languages[index],
                            target_id=row[columns.index('target ID')],
                        )
                    )

        for index, object in enumerate(objects):
            if automatch_target_ids:
                ids_to_replace = Translation._automatch_target_ids(
                    connection=connection,
                    translations=operations[index],
                    id=object,
                    object_type=object_types[index],
                    project_language=project_language,
                    project_id=project_id,
                )
                if ids_to_replace:
                    if config.verbose:
                        logger.info(
                            f"Automatched target IDs: {list(ids_to_replace.keys())} "
                            f"for object: {object} with the "
                            f"following replacements: {list(ids_to_replace.values())}"
                        )
                    for operation in operations[index]:
                        if operation.target_id in ids_to_replace:
                            operation.target_id = ids_to_replace[operation.target_id]
                    if delete and operations_delete[index]:
                        for operation in operations_delete[index]:
                            if operation.target_id in ids_to_replace:
                                operation.target_id = ids_to_replace[
                                    operation.target_id
                                ]
            if operations[index]:
                try:
                    Translation.add_translation(
                        connection=connection,
                        id=object,
                        object_type=int(object_types[index]),
                        translations=operations[index],
                        project_id=project_id,
                    )
                except (IServerError, ValueError) as error:
                    if config.verbose:
                        logger.warning(
                            f"Failed on adding translations for object: {object}"
                            f" with the following error: {error}"
                        )
            if delete and operations_delete[index]:
                try:
                    Translation.remove_translation(
                        connection=connection,
                        id=object,
                        object_type=int(object_types[index]),
                        translations=operations_delete[index],
                        project_id=project_id,
                    )
                except (IServerError, ValueError) as error:
                    if config.verbose:
                        logger.warning(
                            f"Failed on deleting translations for object: {object}"
                            f" with the following error: {error}"
                        )

    @staticmethod
    def _automatch_target_ids(
        connection: Connection,
        translations: list[OperationData],
        id: str,
        object_type: str,
        project_language: int,
        project_id: str,
    ) -> dict:
        """Checks whether all target_ids in the operations are present on
            the server and automatches absent ones if possible.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            translations (list[OperationData]): list of operations to check
            id (str): ID of the object
            object_type (str): type of the object
            project_language (int): lcid of the project's default language
            project_id (str): ID of the project

        Returns:
            A dictionary mapping old target_ids to new ones.
        """
        old_to_new = {}
        object_translations = list_translations(
            connection=connection,
            id=id,
            object_type=object_type,
            languages=[project_language],
            project_id=project_id,
        )
        valid_target_ids = [
            transl.translation_target_id for transl in object_translations
        ]
        for translation in translations:
            if (translation.target_language == project_language) and (
                translation.target_id not in valid_target_ids + list(old_to_new.keys())
            ):
                for existing_translation in object_translations:
                    if (
                        existing_translation.translation_values[0].value
                        == translation.value
                    ):
                        old_to_new[translation.target_id] = (
                            existing_translation.translation_target_id
                        )
                        break
        return old_to_new

    @staticmethod
    def _get_valid_target_ids(
        connection: Connection,
        translations: list[OperationData],
        id: str,
        object_type: str,
        project_id: str,
        operation_type: str,
    ) -> list[OperationData]:
        """Returns a list of target IDs that are valid for the given object.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            translations (list[OperationData]): list of operations to check
            id (str): ID of the object
            object_type (str): type of the object
            project_id (str): ID of the project
            operation_type (str): type of the operation

        Returns:
            List of operations on valid target IDs only.
        """
        new_translations = []
        valid_target_ids = [
            transl.translation_target_id
            for transl in list_translations(
                connection=connection,
                id=id,
                object_type=object_type,
                project_id=project_id,
            )
        ]
        invalid_target_ids = []

        for translation in translations:
            if translation.target_id not in valid_target_ids:
                if translation.target_id not in invalid_target_ids:
                    invalid_target_ids.append(translation.target_id)
                    if config.verbose:
                        logger.warning(
                            f"Target ID: {translation.target_id} is not valid for "
                            f"object: {id}. Skipping {operation_type} translations "
                            "for this target_id."
                        )
            else:
                new_translations.append(translation)
        return new_translations



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.connection import Connection
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.utils.entity import DeleteMixin, Entity
from mstrio.utils.helper import Dictable, delete_none_values
from mstrio.utils.response_processors import languages, objects
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.users_and_groups.user import User

logger = logging.getLogger(__name__)


@method_version_handler(version='11.3.1060')
def list_languages(
    connection: Connection,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list[Language] | list[dict]:
    """Get all languages as a list of Language objects or dictionaries.
    Optionally filter the languages by specifying filters.

    Args:
        connection: Strategy One connection object
        to_dictionary: if True returns a list of Language dicts,
            otherwise returns a list of Language objects
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters:
            ['name', 'base_language_lcid', 'lcid', 'interface_language',
            'last_modified', 'hidden', 'is_language_supported']
    """
    return Language._list_languages(
        connection=connection, to_dictionary=to_dictionary, limit=limit, **filters
    )


@method_version_handler(version='11.3.1060')
def list_interface_languages(
    connection: Connection, to_dictionary: bool = False
) -> list[Language.InterfaceLanguage] | list[dict]:
    """List all available interface languages.

    Args:
        connection (Connection): Strategy One connection object returned
            by `connection.Connection()`
        to_dictionary: if True returns a list of InterfaceLanguage dicts,
            otherwise returns a list of InterfaceLanguage objects

    Returns:
    A list of interface languages as dictionaries."""
    interface_languages = languages.get_interface_languages(connection=connection)
    if to_dictionary:
        return interface_languages
    else:
        return [
            Language.InterfaceLanguage.from_dict(source=obj)
            for obj in interface_languages
        ]


@class_version_handler(version='11.3.1060')
class Language(Entity, DeleteMixin):
    """Python representation of a Strategy One Language object.

    Attributes:
        id (str): language's ID
        name (str): language's name
        base_language_lcid (int): LC ID of the language serving as a base for
            the language
        lcid (int): LC ID of the language
        owner (User): owner of the language
        last_modified (datetime): date of when language was last modified
        formatting_settings (TimeInterval): formatting settings of the language
        interface_language (InterfaceLanguage): details of the language's
            interface language
        hidden (bool): whether the language is hidden
        is_language_supported (bool): whether the language is supported
    """

    @dataclass
    class TimeInterval(Dictable):
        """Object that stores formatting settings of the Language.

        Attributes:
            minutes15 (str): formatting settings for 15 minutes
            minutes30 (str): formatting settings for 30 minutes
            hour (str): formatting settings for an hour
            day (str): formatting settings for a day
            week (str): formatting settings for a week
            hour_of_day (str): formatting settings for the hour of the day
            month (str): formatting settings for a month
            quarter (str): formatting settings for a quarter
            year (str): formatting settings for a year

        Formatting settings instructions:
            To specify formatting for:
                hours: use the letter h
                minutes: use the letter m
                days: use the letter d
                months: use the capital letter M
                quarters: use the letter q
                years: use the letter y
            To specify 12 hour clock use AM/PM afterwards, example: h:mm AM/PM

         Default formatting settings examples:
            minutes15: h:mm AM/PM -> 9:44 AM
            minutes30: h:mm AM/PM -> 9:44 AM
            hour: h:mm AM/PM -> 9:44 AM
            day: M/d -> 7/26
            week: M/d -> 7/26
            hour_of_day: h:mm AM/PM -> 9:44 AM
            month: MMM -> Jul
            quarter: qqq -> Q3
            year: yyyy -> 2023"""

        minutes15: str
        minutes30: str
        hour: str
        day: str
        week: str
        hour_of_day: str
        month: str
        quarter: str
        year: str

    @dataclass
    class InterfaceLanguage(Dictable):
        """Object that stores the interface language of the Language.

        Attributes:
            name (str): name of the interface language
            id (str): id of the interface language
            sub_type (ObjectSubTypes): subtype of the interface language,
                defaults to None
        """

        name: str
        id: str
        sub_type: ObjectSubTypes | None = None

    _OBJECT_TYPE = ObjectTypes.LOCALE
    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'base_language_lcid',
            'lcid',
            'interface_language',
            'last_modified',
            'hidden',
            'is_language_supported',
        ): languages.get,
        (
            'abbreviation',
            'type',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
        ): objects.get_info,
    }
    _API_PATCH = {
        ('name', 'comments', 'owner'): (objects.update, 'partial_put'),
    }

    @staticmethod
    def _parse_owner(source, connection, to_snake_case: bool = True):
        """Parses owner from the API response."""
        from mstrio.users_and_groups import User

        return User.from_dict(source, connection, to_snake_case)

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': _parse_owner,
        'formatting_settings': TimeInterval.from_dict,
        'interface_language': InterfaceLanguage.from_dict,
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
    ) -> None:
        """Initializes a new instance of a Language class

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            id (str, Optional): Language's ID, defaults to None
            name (str, Optional): Language's name, defaults to None

        Note:
            Parameter `name` is not used when fetching. If only `name` parameter
            is provided, `id` will be found automatically if such object exists.

        Raises:
            ValueError: if both `id` and `name` are not provided
                or if Language with the given `name` doesn't exist.
        """
        if not id:
            if name:
                language = self.__find_language_by_name(
                    connection=connection, name=name
                )
                id = language['id']
            else:
                raise ValueError("Please provide either 'id' or 'name' argument.")
        super().__init__(connection=connection, object_id=id, name=name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.name = kwargs.get('name')
        self.base_language_lcid = kwargs.get('base_laguage_lcid')
        self.lcid = kwargs.get('lcid')
        self.interface_language = (
            Language.InterfaceLanguage.from_dict(inl)
            if (inl := kwargs.get('interface_language'))
            else None
        )
        self.last_modified = kwargs.get('last_modified')
        self._hidden = kwargs.get('hidden')
        self._formatting_settings = None
        self.is_language_supported = kwargs.get('is_language_supported')

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        base_language: Language | str | int,
        interface_language_id: str | None = None,
        formatting_settings: TimeInterval | None = None,
    ) -> Language:
        """Create a new language with specified properties.

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            name (str): the name for the new Language
            base_language (Language | str | int): one of the following:
                - lcid attribute of the language that will be used as a base
                    language for the new Language
                - ID of the language that will be used as a base language for
                    the new Language
                - Language class object that will be used as a base language for
                    the new Language
            interface_language_id (str, Optional): id of the new Language's
                interface language
            formatting_settings (TimeInterval, Optional): formatting settings
                for the new Language

        Returns:
            Language class object.
        """
        if isinstance(base_language, int):
            base_lcid = base_language
        elif isinstance(base_language, str):
            base_lcid = Language(connection=connection, id=base_language).lcid
        elif isinstance(base_language, Language):
            base_lcid = base_language.lcid
        else:
            raise ValueError("Please provide a valid base language.")
        body = {
            'name': name,
            'baseLanguageLcid': base_lcid,
            'interfaceLanguageId': interface_language_id,
        }
        body = delete_none_values(source=body, recursion=True)
        language = cls.from_dict(
            source=languages.create(connection=connection, body=body),
            connection=connection,
        )
        if config.verbose:
            logger.info(
                f"Successfully created Language named: '{name}' with ID:"
                f" '{language.id}'"
            )
        if formatting_settings:
            language._formatting_settings = language._update_formatting_settings(
                formatting_settings=formatting_settings
            )
        else:
            language._formatting_settings = Language.TimeInterval.from_dict(
                source=languages.get_formatting_settings(
                    connection=connection, id=language.id
                )
            )
        return language

    def alter(
        self,
        name: str | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
        formatting_settings: TimeInterval | None = None,
    ) -> None:
        """Alter the language's specified properties.

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            name (str, Optional): new name for the Language
            comments (str, optional): long description of the Language
            owner: (str, User, optional): owner of the Language object
            formatting_settings (TimeInverval, Optional): new formatting
                settings for the Language"""

        from mstrio.users_and_groups.user import User

        if isinstance(owner, User):
            owner = owner.id

        properties = {
            'name': name,
            'comments': comments,
            'owner': owner,
        }
        properties = delete_none_values(properties, recursion=False)
        self._alter_properties(**properties)
        if formatting_settings:
            self._formatting_settings = self._update_formatting_settings(
                formatting_settings=formatting_settings
            )

    def _update_formatting_settings(
        self, formatting_settings: TimeInterval
    ) -> TimeInterval:
        """Updates the formatting settings of the Language.

        Args:
            formatting_settings (TimeInterval): new formatting settings

        Returns:
            New formatting settings as TimeInterval class object."""
        formatting_body = {
            'operationList': [
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/minutes15',
                    'value': formatting_settings.minutes15,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/minutes30',
                    'value': formatting_settings.minutes30,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/hour',
                    'value': formatting_settings.hour,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/day',
                    'value': formatting_settings.day,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/week',
                    'value': formatting_settings.week,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/hourOfDay',
                    'value': formatting_settings.hour_of_day,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/month',
                    'value': formatting_settings.month,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/quarter',
                    'value': formatting_settings.quarter,
                },
                {
                    'op': 'replace',
                    'path': '/formattingSettings/timeInterval/year',
                    'value': formatting_settings.year,
                },
            ]
        }
        return Language.TimeInterval.from_dict(
            languages.update_formatting_settings(
                connection=self.connection, id=self.id, body=formatting_body
            )
        )

    def _remove_interface_language(self) -> None:
        """Removes the interface language from the Language if one exists."""
        if self.interface_language:
            op_list = [
                {
                    'op': 'remove',
                    'path': '/interfaceLanguage',
                    'value': self.interface_language.id,
                }
            ]
            languages.update(
                connection=self.connection, id=self.id, body={'operationList': op_list}
            )
            self.interface_language = None

    @classmethod
    def _list_languages(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        limit: int | None = None,
        **filters,
    ) -> list[Language] | list[dict]:
        objects = languages.get_all(connection=connection, limit=limit, filters=filters)
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @staticmethod
    def __find_language_by_name(connection: Connection, name: str):
        languages = list_languages(connection=connection, name=name)

        if languages:
            number_of_languages = len(languages)
            if number_of_languages > 1:
                raise ValueError(
                    f"There are {number_of_languages} Languages"
                    " with this name. Please initialize with id."
                )
            else:
                return languages[0].to_dict()
        else:
            raise ValueError(f"There is no Language with the given name: '{name}'")

    @property
    def hidden(self):
        return self._hidden

    @property
    def formatting_settings(self):
        if self._formatting_settings:
            return self._formatting_settings
        else:
            self._formatting_settings = Language.TimeInterval.from_dict(
                source=languages.get_formatting_settings(
                    connection=self.connection, id=self.id
                )
            )



import logging
import math
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

import pandas as pd
from tqdm.auto import tqdm

from mstrio import config
from mstrio.api import cubes, datasets
from mstrio.connection import Connection
from mstrio.object_management import (
    Folder,
    PredefinedFolders,
    get_predefined_folder_contents,
)
from mstrio.object_management.search_operations import SearchPattern, full_search
from mstrio.utils import helper
from mstrio.utils.encoder import Encoder
from mstrio.utils.entity import CertifyMixin, ObjectSubTypes
from mstrio.utils.model import Model
from mstrio.utils.resolvers import (
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.version_helper import is_server_min_version

from .cube import _Cube

if TYPE_CHECKING:
    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


def list_super_cubes(
    connection: Connection,
    name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["SuperCube"] | list[dict]:
    """Get list of SuperCube objects or dicts with them.
    Optionally filter cubes by specifying 'name'.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name = ?onny will return Sonny and Tonny

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of super cubes being searched
        search_pattern (SearchPattern enum or int, optional): pattern to search
            for, such as Begin With or Contains. Possible values are available
            in ENUM mstrio.object_management.SearchPattern.
            Default value is BEGIN WITH (4).
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns SuperCube objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        **filters: Available filter parameters: ['id', 'name', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'owner',
            'ext_type', 'view_media', 'certified_info']

    Returns:
        list with SuperCubes or list of dictionaries
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    validate_owner_key_in_filters(filters)

    objects_ = full_search(
        connection,
        object_types=ObjectSubTypes.SUPER_CUBE,
        project=proj_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        **filters,
    )
    if to_dictionary:
        return objects_
    return [
        SuperCube.from_dict(source=obj_, connection=connection, with_missing_value=True)
        for obj_ in objects_
    ]


@dataclass
class SuperCubeFormExpression:
    """A class representing a form expression used to define multiform
    attributes for Super Cubes. Links an attribute form to the column
    in the table.

    Attributes:
        table (str): name of a table
        column (str): name of a column
    """

    table: str
    column: str

    def to_dict(self):
        return {'tableName': self.table, 'columnName': self.column}


@dataclass
class SuperCubeAttributeForm:
    """A class representing an attribute form for used to define multiform
    attributes for Super Cubes.

    Attributes:
        category (str): form's category
        expressions (list[SuperCubeFormExpressions]): list of form expressions
    """

    category: str
    expressions: list[SuperCubeFormExpression]

    def to_dict(self):
        return {
            'category': self.category,
            'expressions': [expr.to_dict() for expr in self.expressions],
        }


@dataclass
class SuperCubeAttribute:
    """A class representing a multiform attribute for Super Cubes.

    Attributes:
        name (str): a name of multiform attribute
        forms (list[SuperCubeAttributeForm]): a list of attribute forms
    """

    name: str
    forms: list[SuperCubeAttributeForm]

    def to_dict(self):
        return {
            'name': self.name,
            'attributeForms': [form.to_dict() for form in self.forms],
        }


class SuperCube(_Cube, CertifyMixin):
    """Manage multiple table cube (MTDI aka Super Cube) - according to
    EnumDSSXMLObjectSubTypes its subtype is 779 (DssXmlSubTypeReportEmmaCube).
    It inherits all properties from Cube.

    Attributes:
        connection: Strategy One connection object returned by
            `connection.Connection()`.
        id: Identifier of a pre-existing cube.
        instance_id (str): Identifier of a cube instance if already initialized,
            None by default.
        name: Name of the SuperCube.
        description: Description given to the SuperCube.
        id: Unique identifier for the super cube. Used to update a
            existing super cube or generated after creating a new super cube.
            (deprecated)
        upload_body (string): upload body of super cube
        session_id (string): ID of session used for uploading and publishing
            super cube
        size(integer): size of cube
        status(integer): status of cube
        path(string): full path of the cube on environment
        owner_id(string): ID of cube's owner
        attributes(list): all attributes of cube
        metrics(list): all metrics of cube
        attr_elements(list): all attributes elements of cube
        selected_attributes(list): IDs of filtered attributes
        selected_metrics(list): IDs of filtered metrics
        selected_attr_elements(list): IDs of filtered attribute elements
        dataframe(object): content of a cube extracted into a
            Pandas `DataFrame`
        table_definition
    """

    _OBJECT_SUBTYPE = ObjectSubTypes.SUPER_CUBE.value
    __VALID_POLICY = ['add', 'update', 'replace', 'upsert']
    __MAX_DESC_LEN = 250

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
        description: str | None = None,
        instance_id: str | None = None,
        progress_bar: bool = True,
        parallel: bool = True,
    ):
        """Initialize super cube.

        When creating new super cube, provide its `name` and an optional
        `description`.
        When updating a pre-existing super cube, provide its `id`. This
        identifier will be then used to initialize super cube.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            id (str, optional): Identifier of a pre-existing super cube.
                Used when updating a pre-existing super cube.
            name (str): Name of the super cube.
            description (str, optional): Description of the super cube. Must be
                less than or equal to 250 characters.
            instance_id (str): Identifier of an instance if cube instance has
                been already initialized, NULL by default.
            progress_bar(bool, optional): If True (default), show the upload
                progress bar.
            parallel (bool, optional): If True (default), utilize optimal number
                of threads to increase the download speed. If False, this
                feature will be disabled.
        """
        if name is not None:
            self.__check_param_str(name, msg="Super cube name should be a string.")
            self.__check_param_len(
                name,
                msg=f"Super cube name should be <= {self.__MAX_DESC_LEN} characters.",
                max_length=self.__MAX_DESC_LEN,
            )

        if description is not None:
            self.__check_param_str(
                description, msg="Super cube description should be a string."
            )
            self.__check_param_len(
                description,
                msg="Super cube description should be <= {} characters.".format(
                    self.__MAX_DESC_LEN
                ),
                max_length=self.__MAX_DESC_LEN,
            )

        connection._validate_project_selected()

        if id is not None:
            super().__init__(
                connection=connection,
                id=id,
                instance_id=instance_id,
                parallel=parallel,
                progress_bar=progress_bar,
            )
            if config.verbose and self._subtype != ObjectSubTypes.SUPER_CUBE.value:
                logger.warning(
                    f"Warning: Object type mismatch. Object with provided "
                    f"ID: '{id}' is not a SuperCube but an instance of: "
                    f"'{ObjectSubTypes(self._subtype).name}'."
                )
        else:
            self._init_variables(
                connection=connection,
                name=name,
                description=description,
                progress_bar=progress_bar,
                parallel=parallel,
            )

    def _init_variables(self, default_value=None, **kwargs):
        super()._init_variables(default_value=default_value, **kwargs)
        self._tables = []
        self._session_id = None
        # used to check publish status after completing publish
        self.__last_session_id = None
        self._folder_id = None
        self.__upload_body = None
        # used to store indexes for every table after an update to have correct
        # index in case of doing multiple updates without publish
        self.__update_indexes = {}

    def add_table(
        self, name, data_frame, update_policy, to_metric=None, to_attribute=None
    ):
        """Add a `Pandas.DataFrame` to a collection of tables which are later
        used to populate the Strategy One super cube with data.

        Args:
            name (str): Logical name of the table that is visible to users of
                the super cube in Strategy One.
            data_frame (:obj:`pandas.DataFrame`): Pandas DataFrame to add or
                update.
            update_policy (str): Update operation to perform. One of 'add'
                (inserts new, unique rows), 'update' (updates data in existing
                rows and columns), 'upsert' (updates existing data and inserts
                new rows), or 'replace' (replaces the existing data with new
                data).
            to_metric (optional, :obj:`list` of str): By default, Python numeric
                data types are treated as metrics while character and date types
                are treated as attributes. For example, a column of integer-like
                strings ("1", "2", "3") would, by default, be an attribute. If
                the intent is to format this data as a metric, provide the
                respective column name as a string in a list to the `to_metric`
                parameter.
            to_attribute (optional, :obj:`list` of str): Logical opposite of
                `to_metric`. Helpful for formatting an integer-based row
                identifier as a primary key in the super cube.
        """
        update_policy = update_policy.lower()

        if not isinstance(data_frame, pd.DataFrame):
            raise TypeError(
                "`data_frame` parameter must be a valid `Pandas.DataFrame`."
            )
        if not helper.check_duplicated_column_names(data_frame):
            raise ValueError(
                "`DataFrame` column names need to be unique for each table in the super"
                " cube."
            )
        self.__check_update_policy(update_policy)
        if to_attribute and to_metric and any(col in to_attribute for col in to_metric):
            raise ValueError(
                "Column name(s) present in `to_attribute` also present in `to_metric`."
            )

        table = {
            "table_name": name,
            "data_frame": data_frame,
            "update_policy": update_policy,
        }

        if to_attribute is not None:
            if any(col for col in to_attribute if col not in data_frame.columns):
                raise ValueError(
                    "Column name(s) in `to_attribute` were not found in"
                    " `DataFrame.columns`."
                )
            else:
                table["to_attribute"] = to_attribute

        if to_metric is not None:
            if any(col for col in to_metric if col not in data_frame.columns):
                raise ValueError(
                    "Column name(s) in `to_metric` were not found in"
                    " `DataFrame.columns`."
                )
            else:
                table["to_metric"] = to_metric

        self._tables.append(table)
        if not self.__update_indexes.get(name):
            self.__update_indexes[name] = 0

    def __check_update_policy(self, update_policy: str) -> None:
        if update_policy not in self.__VALID_POLICY:
            raise ValueError(f"Update policy must be one of {self.__VALID_POLICY}.")

    def remove_table(self, name):
        """Removes a table from a collection of tables which are
        later used to populate the Strategy One super cube with data.

        Note: this operation is executed locally and is used only to prepare
            data before sending it to server. You can check current state of
            tables with property `tables`.

        Args:
            name (str): Logical name of the table that is visible to users of
                the super cube in Strategy One.
        """
        self._tables = [t for t in self._tables if t.get('table_name') != name]

    def __check_folder_contents(self, name: str, folder_id: str | None = None):
        """Check if folder already have object with given name.
        If folder_id is None, check My Reports folder.
        """

        if folder_id:
            folder = Folder(self.connection, id=folder_id)
            contents = folder.get_contents(name=name)
        else:
            contents = get_predefined_folder_contents(
                self.connection, PredefinedFolders.PROFILE_REPORTS, name=name
            )

        if contents:
            raise ValueError(
                f"Super Cube with name {self.name} already exist in"
                f" {folder_id or 'My Reports'} folder. If you want to override already"
                " existing cube, add tables with update policy `replace` and call"
                " 'create()' method with argument 'force=True'."
            )

    def __are_all_tables_with_replace_policy(self):
        return all(table['update_policy'] == 'replace' for table in self.tables)

    def create(
        self,
        folder_id: str | None = None,
        auto_upload: bool = True,
        auto_publish: bool = True,
        chunksize: int = 100000,
        force: bool = False,
        attribute_forms: list[SuperCubeAttribute] | None = None,
    ) -> None:
        """Create a new super cube and initialize cube object after successful
        creation. This function does not return new super cube, but it updates
        object inplace.

        Args:
            folder_id (str, optional): ID of the shared folder in which the
                super cube will be created. If `None`, defaults to the user's
                My Reports folder.
            auto_upload (bool, optional): If True, automatically uploads the
                data to the I-Server. If False, simply creates the super cube
                definition but does not upload data to it.
            auto_publish (bool, optional): If True, automatically publishes the
                data used to create the super cube definition. If False, simply
                creates the super cube but does not publish it. To publish the
                super cube, data has to be uploaded first.
            chunksize (int, optional): Number of rows to transmit to the
                I-Server with each request when uploading.
            force (bool, optional): If True, skip checking if a super cube
                already exist in the folder with the given name.
                Defaults to False.
            attribute_forms (list[SuperCubeAttribute], optional): list
            of instances of SuperCubeAttribute, that links tables' columns to
            different forms of the same attribute. If this argument is not
            provided, a separate attribute is created for every distinct column
            name. If it is provided, it contains mapping of columns to
            attributes as different form of an attribute. For all columns that
            are not part of any attribute's form expression, separate attribute
            will be created.
        """
        if auto_publish and not auto_upload:
            raise ValueError(
                "Data needs to be uploaded to the I-Server before the super cube can be"
                " published."
            )

        if folder_id is not None:
            self._folder_id = folder_id
        else:
            self._folder_id = ""

        if not force:
            self.__check_folder_contents(self.name, folder_id)

        if (
            force
            and is_server_min_version(self.connection, '11.2.0300')
            and not self.__are_all_tables_with_replace_policy()
        ):
            raise ValueError(
                "All the tables must be added with update policy 'replace', when trying"
                " to override existing Super Cube with 'force=True'."
            )

        # generate model of the super cube
        self.__build_model(attribute_forms)

        # makes request to create the super cube
        response = datasets.create_multitable_dataset(
            connection=self._connection, body=self.__model
        )
        self._set_object_attributes(**response.json())

        if config.verbose:
            logger.info(
                "Created super cube '{}' with ID: '{}'.".format(*[self.name, self._id])
            )

        if auto_upload:
            self.update(chunksize=chunksize, auto_publish=auto_publish)

        # after creating super cube fetch definition and create filter object
        self._get_definition()

    def update(self, chunksize: int = 100000, auto_publish: bool = True):
        """Updates a super cube with new data.

        Args:
            chunksize (int, optional): Number of rows to transmit to the server
                with each request.
            auto_publish: If True, automatically publishes the data used to
                update the super cube definition to the super cube. If False,
                simply updates the super cube but does not publish it.
        """

        # form request body and create a session for data uploads
        self.__form_upload_body()

        if not self._session_id:
            response = datasets.upload_session(
                connection=self._connection, id=self._id, body=self.__upload_body
            )
            response_json = response.json()
            self._session_id = response_json['uploadSessionId']
            self.__last_session_id = None

        # upload each table
        for ix, _table in enumerate(self._tables):
            _df, _name = _table["data_frame"], _table["table_name"]

            # break the data up into chunks using a generator
            chunks = (_df[i : i + chunksize] for i in range(0, _df.shape[0], chunksize))

            total = _df.shape[0]

            # Count the number of iterations
            it_total = math.ceil(total / chunksize)

            pbar = tqdm(chunks, total=it_total, disable=(not self._progress_bar))
            for index, chunk in enumerate(pbar):
                pbar.set_description(f"Uploading {ix + 1}/{len(self._tables)}")

                # if the chunk is empty we skip uploading it entirely
                if not chunk.empty:
                    # base64 encode the data
                    encoder = Encoder(data_frame=chunk, dataset_type='multi')
                    b64_enc = encoder.b64_data

                    # form body of the request
                    body = {
                        "tableName": _name,
                        "index": self.__update_indexes.get(_name, 0) + index + 1,
                        "data": b64_enc,
                    }

                    # make request to upload the data
                    response = datasets.upload(
                        connection=self._connection,
                        id=self._id,
                        session_id=self._session_id,
                        body=body,
                        throw_error=False,
                    )

                    if not response.ok:
                        # on error, cancel the previously uploaded data
                        datasets.publish_cancel(
                            connection=self._connection,
                            id=self._id,
                            session_id=self._session_id,
                        )
                        self.reset_session()
                        pbar.close()
                        return

                pbar.set_postfix(rows=min((index + 1) * chunksize, total))
            pbar.close()

            # prepare index in case of the next update operation for this table
            # without publishing
            self.__update_indexes[_name] += it_total
        self._tables = []

        # if desired, automatically publish the data to the new super cube
        if auto_publish:
            self.publish()

    def save_as(
        self,
        name: str,
        description: str | None = None,
        folder_id: str | None = None,
        table_name: str | None = None,
    ) -> "SuperCube":
        """Creates a new single-table cube with the data frame stored in the
        SuperCube instance `SuperCube.dataframe`.

        Args:
            name(str): Name of cube.
            description(str): Description of the cube.
            folder_id (str, optional): ID of the shared folder that the super
                cube should be created within. If `None`, defaults to the user's
                My Reports folder.
            table_name (str, optional): Name of the table. If None (default),
                the first table name of the original cube will be used.
        """
        if len(self._table_names) > 1:
            helper.exception_handler(
                msg="This feature works only for the single-table cubes. "
                "To export multi-table cube use `create` method."
            )
        elif self._dataframe is None:
            raise ValueError(
                "Current cube has to be fetched with `to_dataframe` method first to"
                " create a copy."
            )
        else:
            if table_name is None:
                table_name = self._table_names[0]["name"]
            new_cube = SuperCube(self._connection, name=name, description=description)
            new_cube.add_table(
                name=table_name, data_frame=self.dataframe, update_policy="replace"
            )
            new_cube.create(folder_id=folder_id)
            return new_cube

    def publish(self) -> bool:
        """Publish the uploaded data to the selected super cube.

        Note:
            A super cube can be published just once.

        Returns:
            True if the data was published successfully, else False.
        """
        response = datasets.publish(
            connection=self._connection, id=self._id, session_id=self._session_id
        )

        if not response.ok:
            # on error, cancel the previously uploaded data
            datasets.publish_cancel(
                connection=self._connection, id=self._id, session_id=self._session_id
            )
            self.reset_session()
            return False

        status = 6  # default initial status
        while status != 1:
            status = self.publish_status()['status']
            time.sleep(1)
            if status == 1:
                # clear instance_id to force new instance creation
                self.instance_id = None
                self.reset_session()
                if config.verbose:
                    logger.info(f"Super cube '{self.name}' published successfully.")
                return True

    def publish_status(self):
        """Check the status of data that was uploaded to a super cube.

        Returns: status: The status of the publication process as a dictionary.
            In the 'status' key, "1" denotes completion.
        """
        # after publish, `self._session_id` is reset to force new session
        # creation in the next update, so we have to use its value saved in
        # `self.__last_session_id`
        session_id = self._session_id if self._session_id else self.__last_session_id

        return self.upload_status(
            connection=self.connection, id=self.id, session_id=session_id
        )

    def reset_session(self):
        """Reset upload session."""
        if self._session_id:
            # save last session id in case of checking status of publish
            self.__last_session_id = self._session_id
            # clear session_id to force new session creation
            self._session_id = None
            self.__update_indexes = {}

    def export_sql_view(self) -> str:
        """Export SQL View of a Super Cube.

        Returns:
            SQL View of a Super Cube.
        """
        res = cubes.get_sql_view(self._connection, self._id)
        sql_statement = res.json()['sqlStatement']
        return sql_statement

    def __build_model(self, attr_forms_mapping: list[SuperCubeAttribute] = None):
        """Create json representation of the super cube."""

        # generate model
        model = Model(
            tables=self._tables,
            name=self.name,
            description=self.description,
            folder_id=self._folder_id,
            attr_forms_mapping=attr_forms_mapping,
        )
        self.__model = model.get_model()

    def __form_upload_body(self):
        """Form request body for creating an upload session for data
        uploads."""

        # generate body string
        body = {
            "tables": [
                {
                    "name": tbl["table_name"],
                    "updatePolicy": tbl["update_policy"],
                    "columnHeaders": list(tbl["data_frame"].columns),
                }
                for tbl in self._tables
            ]
        }
        self.__upload_body = body

    @staticmethod
    def upload_status(connection: Connection, id: str, session_id: str):
        """Check the status of data that was uploaded to a super cube.

        Args:
            connection (Optional[Connection]): Strategy One connection object
                returned by `connection.Connection()`.
            id (Optional[str]): Identifier of a pre-existing super cube.
            session_id (Optional[str]): Identifier of the server session used
                for collecting uploaded data.

        Returns:
            status: The status of the publication process as a dictionary. In
                the 'status' key, "1" denotes completion.
        """
        if not session_id:
            raise AttributeError("No upload session created.")
        else:
            response = datasets.publish_status(
                connection=connection, id=id, session_id=session_id
            )
            return response.json()

    @staticmethod
    def __check_param_len(param, msg, max_length):
        if len(param) > max_length:
            raise ValueError(msg)
        else:
            return True

    @staticmethod
    def __check_param_str(param, msg):
        if not isinstance(param, str):
            raise TypeError(msg)
        else:
            return True

    @property
    def upload_body(self):
        return self.__upload_body

    @property
    def session_id(self) -> str:
        return self._session_id

    @property
    def tables(self):
        return self._tables



import logging
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import monitors
from mstrio.server.cluster import Cluster
from mstrio.utils.cache import Cache
from mstrio.utils.helper import camel_to_snake, exception_handler, fetch_objects_async
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0000')
def list_cube_caches(
    connection: "Connection",
    nodes: list[str] | str | None = None,
    cube_id: str | None = None,
    loaded: bool | None = False,
    db_connection_id: str | None = None,
    project_ids: list[str] | None = None,
    to_dictionary: bool | None = False,
    limit: int | None = None,
) -> list["CubeCache"] | list[dict]:
    """List cube caches. You can filter them by cube (`cube_id`), database
    connection (`db_connection_id`) and projects (`project_ids`). You can also
    obtain only loaded caches (`loaded=True`).

    You can specify from which `nodes` caches will be retrieved. If `nodes` are
    `None` then all nodes are retrieved from the cluster.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`.
        nodes (list of strings or string, optional): names of nodes on which
            caches will be searched. By default, it equals `None` and in that
            case all nodes' names are loaded from the cluster.
        cube_id (string, optional): When provided, only caches for the cube with
            given ID will be returned (if any).
        loaded (bool, optional): If True then only loaded caches will be
            retrieved. Otherwise, all cubes will be returned.
        db_connection_id (string, optional): When provided, only caches for the
            database connection with given ID will be returned (if any).
        project_ids (list of string, optional): When provided only caches
            for projects with given IDs will be returned (if any).
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns CubeCache objects
        limit(integer, optional): Cut-off value for the number of objects
            returned. Default value is `None` which means no limit.

    Returns:
        List of CubeCache objects when parameter `to_dictionary` is set to False
        (default value) or list of dictionaries otherwise.
    """
    if project_ids is not None:
        project_ids = ','.join(project_ids)  # form accepted by request

    if nodes is None:
        cluster_ = Cluster(connection)
        nodes = cluster_.list_nodes(project=connection.project_id, to_dictionary=True)
        nodes = [node.get('name') for node in nodes]

    nodes = [nodes] if isinstance(nodes, str) else nodes
    caches = []
    for node in nodes:
        caches += fetch_objects_async(
            connection=connection,
            api=monitors.get_cube_caches,
            async_api=monitors.get_cube_caches_async,
            dict_unpack_value='cubeCaches',
            node=node,
            limit=limit,
            project_ids=project_ids,
            chunk_size=1000,
            loaded=loaded,
            filters={},
        )
    if cube_id:
        caches = [
            cache
            for cache in caches
            if cache.get('source', {}).get('id', '') == cube_id
        ]
    if db_connection_id:
        caches = [
            cache
            for cache in caches
            if db_connection_id
            in [db.get('id', '') for db in cache.get('databaseConnections', [])]
        ]
    if to_dictionary:
        return caches
    return CubeCache.from_dict(connection, caches)


def delete_cube_caches(
    connection: "Connection",
    loaded: bool = False,
    force: bool = False,
    nodes: list[str] | str | None = None,
    cube_id: str | None = None,
    db_connection_id: str | None = None,
) -> dict | None:
    """Delete all cube caches on a given node.

    Optionally it is possible to specify for which cube or for which database
    connection caches will be deleted. It is also possible to delete only loaded
    caches.

    You can specify from which `nodes` caches will be deleted. If `nodes` are
    `None` then all nodes are retrieved from cluster.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`.
        nodes (list of strings or string, optional): names of nodes from which
            caches will be deleted. By default it equals `None` and in that
            case all nodes names are loaded from the cluster.
        cube_id (string, optional): When provided, only caches for the cube with
            given ID will be deleted (if any).
        db_connection_id (string, optional): When provided, only caches for the
            database connection with given ID will be deleted (if any).
        loaded (bool, optional): If True then only loaded caches will be
            deleted. Otherwise all cubes will be returned.
        force (bool, optional): If True, then no additional prompt will be shown
            before deleting caches for given cube. Default is False.

    Returns:
        Dictionary with two keys:
         - 'succeeded' - list with IDs of caches which were deleted
         - 'failed' - list with IDS of caches which were not deleted
        or None when user doesn't agree for deletion when prompted
    """
    user_input = 'N'
    if not force:
        # construct meaningful message for deletion input
        tmp_msg = ['all']
        if loaded is True:
            tmp_msg.append('loaded')
        tmp_msg.append('caches')
        if cube_id:
            tmp_msg.append(f"for cube with ID: '{cube_id}'")
        if db_connection_id:
            tmp_msg.append(f"from database connection with ID: '{db_connection_id}'")
        msg = f"Are you sure you want to delete {' '.join(tmp_msg)}? [Y/N]:"
        user_input = input(msg) or 'N'
    if force or user_input == 'Y':
        caches = list_cube_caches(connection, nodes, cube_id, loaded, db_connection_id)
        succeeded = []
        failed = []
        for cache in caches:
            if cache.delete(force=True):
                succeeded.append(cache.id)
            else:
                failed.append(cache.id)
        return {'succeeded': succeeded, 'failed': failed}
    return None


def delete_cube_cache(connection: "Connection", id: str, force: bool = False):
    """Delete single cube cache.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`.
        id (string): cube cache ID
        force (bool, optional): If True, then no additional prompt will be shown
            before deleting caches for given cube. Default is False.

    Returns:
        True for success. False otherwise.
    """
    return CubeCache._delete(connection, id, force)


@class_version_handler('11.3.0000')
class CubeCache(Cache):
    """
    Manage cube cache.
    """

    def __init__(
        self,
        connection: "Connection",
        cache_id: str,
        cube_cache_dict: dict | None = None,
    ):
        """Initialize the CubeCache object. If cube_cache_dict provided
        no I-Server request will be sent.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            cache_id (string): cube cache id
            cube_cache_dict (dict, optional): dictionary with properties of cube
                cache object. If it is provided then cache dictionary will not
                be retrieved from I-Server.
        """
        super().__init__(connection, cache_id)

        if not cube_cache_dict:
            self.fetch()
        else:
            self._init_variables(**cube_cache_dict)

    def _init_variables(self, **cube_cache_dict):
        cube_cache_dict = camel_to_snake(cube_cache_dict)
        super()._init_variables(**cube_cache_dict)
        self._state = cube_cache_dict.get('state')
        self._last_update_job = cube_cache_dict.get('last_update_job')
        self._open_view_count = cube_cache_dict.get('open_view_count')
        self._historic_hit_count = cube_cache_dict.get('historic_hit_count')
        self._database_connections = cube_cache_dict.get('database_connections', [])
        self._file_name = cube_cache_dict.get('file_name')
        self._data_languages = cube_cache_dict.get('data_languages', [])
        self._row_count = cube_cache_dict.get('row_count')
        self._column_count = cube_cache_dict.get('column_count')
        self._creator_id = cube_cache_dict.get('creator_id')
        self._job_execution_statistics = cube_cache_dict.get('job_execution_statistics')

    def fetch(self):
        res = monitors.get_cube_cache_info(self._connection, self._id)
        self._init_variables(**res.json())

    @classmethod
    def from_dict(
        cls, connection: "Connection", caches: list[dict]
    ) -> list["CubeCache"]:
        return [
            CubeCache(connection, cache_dict['id'], cache_dict) for cache_dict in caches
        ]

    def __alter_status(self, active: bool = None, loaded: bool = None) -> str | None:
        if active is not None and loaded is not None:
            msg = 'You cannot set both states during one request'
            exception_handler(msg, UserWarning)
            return None
        res = monitors.alter_cube_cache_status(
            self._connection, self._id, active, loaded, False
        )
        return res.json()['manipulationId']

    def load(self):
        """Load cache."""
        return self.__alter_status(loaded=True)

    def unload(self):
        """Unload cache."""
        return self.__alter_status(loaded=False)

    def activate(self):
        """Activate cube cache."""
        return self.__alter_status(active=True)

    def deactivate(self):
        """Deactivate cube cache."""
        return self.__alter_status(active=False)

    def delete(self, force: bool = False):
        """Delete cube cache.

        Args:
            force: If True, then no additional prompt will be shown before
                deleting CubeCache.

        Returns:
            True for success. False otherwise.
        """
        return self._delete(self._connection, self._id, force)

    @staticmethod
    def _delete(connection: "Connection", id: str, force: bool = False):
        user_input = 'N'
        if not force:
            user_input = (
                input(
                    f"Are you sure you want to delete cube cache"
                    f"with ID: '{id}'? [Y/N]: "
                )
                or 'N'
            )
        if force or user_input == 'Y':
            response = monitors.delete_cube_cache(connection, id, False)
            if response.status_code == 204:
                if config.verbose:
                    logger.info(f"Successfully deleted cube cache with ID: '{id}'.")
                return True
            else:
                return False
        else:
            return False

    def list_properties(self):
        """List properties for cube cache."""
        return {
            **super().list_properties(),
            'row_count': self.row_count,
            'column_count': self.column_count,
            'historic_hit_count': self.historic_hit_count,
            'open_view_count': self.open_view_count,
            'creator_id': self.creator_id,
            'last_update_job': self.last_update_job,
        }

    def get_manipulation_status(self, manipulation_id: str) -> dict:
        """Get manipulation status of cube cache."""
        res = monitors.get_cube_cache_manipulation_status(
            self._connection, manipulation_id, False
        )
        return res.json()['status']

    @property
    def state(self):
        return self._state

    @property
    def database_connections(self):
        return self._database_connections

    @property
    def file_name(self):
        return self._file_name

    @property
    def data_languages(self):
        return self._data_languages

    @property
    def job_execution_statistics(self):
        if self._job_execution_statistics is None:
            self.fetch()
        return self._job_execution_statistics

    @property
    def row_count(self):
        return self._row_count

    @property
    def column_count(self):
        return self._column_count

    @property
    def historic_hit_count(self):
        return self._historic_hit_count

    @property
    def open_view_count(self):
        return self._open_view_count

    @property
    def creator_id(self):
        return self._creator_id

    @property
    def last_update_job(self):
        return self._last_update_job



import logging

from mstrio import config
from mstrio.api import emails
from mstrio.connection import Connection
from mstrio.users_and_groups import User

logger = logging.getLogger(__name__)


def send_email(
    connection: 'Connection',
    users: list[str | User],
    subject: str,
    content: str,
    is_html: bool | None = None,
) -> None:
    """Send an email to specified recipients.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        users(list[str | User]): List of user IDs or User objects to send
            the email to
        subject (str): Subject of the email
        content (str): Content of the email
        is_html (bool, optional): Whether the email content is HTML-formatted
    """
    if not subject or not content:
        raise ValueError("Both `subject` and `content` must be provided.")

    body = {
        'notificationType': 'USER_CREATION',
        'userIds': [
            user_id.id if isinstance(user_id, User) else user_id for user_id in users
        ],
        'subject': subject,
        'content': content,
        'isHTML': is_html,
    }

    emails.send_email(connection=connection, body=body)
    if config.verbose:
        logger.info(f"Email sent with the subject: {subject}")



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error creating a Change Journal search instance.")
def create_change_journal_search_instance(
    connection: Connection, body: dict, fields: str | None = None
) -> Response:
    """Create a Change Journal search instance.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (dict): JSON-formatted body of the search instance
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.post(
        endpoint='/api/changeJournal', json=body, params={'fields': fields}
    )


@ErrorHandler(err_msg="Error getting Change Journal search results")
def get_change_journal_search_results(
    connection: Connection,
    search_id: str,
    offset: int = 0,
    limit: int = -1,
    fields: str | None = None,
) -> Response:
    """Get Change Journal search results.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        search_id (str): ID of the search instance
        offset (int, optional): Starting point within the collection of
            returned search results
        limit (int, optional): Maximum number of items returned for a single
            search request
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.get(
        endpoint=f'/api/changeJournal/{search_id}',
        params={'limit': limit, 'offset': offset, 'fields': fields},
    )



from typing import TYPE_CHECKING

from requests import Response

from mstrio.utils.api_helpers import changeset_manager
from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@ErrorHandler(
    err_msg="Selected project {name} does not exist or is not loaded."
    " Please load the project or select a valid project "
    "or create a new project using `create_new` method"
)
def get_project(
    connection: 'Connection',
    name: str,
    error_msg: str | None = None,
    throw_error: bool = True,
    whitelist: list = [('ERR001', 500), ('ERR014', 403)],  # noqa: B006
) -> Response:
    """Get a specific project that the authenticated user has access to.

    Args:
        connection (Connection): Strategy One REST API connection object
        name (string): Strategy One project name
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list, optional): list of tuples of I-Server Error and
            HTTP errors codes respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """
    return connection.get(endpoint=f'/api/projects/{name}')


@ErrorHandler(err_msg="Error fetching list of available projects.")
def get_projects(
    connection: 'Connection',
    error_msg: str | None = None,
    whitelist: list | None = None,
) -> Response:
    """Get a list of all projects that the authenticated user has access to.

    Args:
        connection (Connection): Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list, optional): list of tuples of I-Server Error and
            HTTP errors codes respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """

    return connection.get(endpoint='/api/projects', headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error creating a new project.")
def create_project(
    connection: 'Connection',
    name: str,
    description: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Create a new project, either synchronously or asynchronously.

    Args:
        connection (Connection): Strategy One REST API connection object
        name (string): Name of the new project
        description (string, optional): Description of the new project
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    body = {'name': name, 'description': description}
    return connection.post(
        endpoint='/api/v2/projects',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error getting project import quota for project with ID {id}")
def get_project_import_quota(
    connection: 'Connection', id: str, error_msg: str | None = None
) -> Response:
    """Get the amount of space, in MB, that can be used for the Data Import
    function for a specific project. This is the default value applied to all
    users.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(endpoint=f'/api/projects/{id}/quotas')


@ErrorHandler(err_msg="Error setting import quota for project with ID {id}")
def set_project_import_quota(
    connection: 'Connection', id: str, body: dict, error_msg: str | None = None
) -> Response:
    """Set the amount of space, in MB, that can be used for the Data Import
    function for a specific project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project id string
        body (dict): JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint=f'/api/projects/{id}/quotas',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(
    err_msg="Error setting user {user_id} import quota for project with ID {id}"
)
def set_user_import_quota(
    connection: 'Connection',
    id: str,
    user_id: str,
    body: dict,
    error_msg: str | None = None,
) -> Response:
    """Set the amount of space, in MB, that can be used for the Data Import
    function for a specific user. The value provided is rounded to an integer.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project id string
        user_id (string): User ID string
        body (dict): JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint=f'/api/projects/{id}/users/{user_id}/quotas',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error getting user import quota for project with ID {id}")
def get_user_import_quota(
    connection: 'Connection',
    id: str,
    user_id: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a list of users for whom Data Import quotas have been set and their
    quotas.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project id string
        user_id (string, optional): User id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/projects/{id}/users/quotas',
        headers={'X-MSTR-ProjectID': None},
        params={'user_id': user_id},
    )


@ErrorHandler(err_msg="Error fetching project {id} settings configuration.")
def get_project_settings_config(
    connection: 'Connection', id: str, error_msg: str | None = None
) -> Response:
    """Get project settings configurations.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/v2/projects/{id}/settings/config',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting project settings for project with ID {id}")
def get_project_settings(
    connection: 'Connection',
    id: str,
    error_msg: str | None = None,
    whitelist: list | None = None,
) -> Response:
    """Get project settings.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list, optional): list of tuples of I-Server Error and
            HTTP errors codes respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/v2/projects/{id}/settings',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error setting project settings for project with ID {id}")
def set_project_settings(
    connection: 'Connection', id: str, body: dict, error_msg: str | None = None
) -> Response:
    """Set new project settings.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string, optional): Project id string
        body (dict): body of the request
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint=f'/api/v2/projects/{id}/settings',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error updating project settings for project with ID {id}")
def update_project_settings(
    connection: 'Connection', id: str, body: dict, error_msg: str | None = None
) -> Response:
    """Update project settings.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project id string
        body (dict): body of the request
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(
        endpoint=f'/api/v2/projects/{id}/settings',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error getting engine setting for project with ID {id}")
def get_engine_settings(
    connection: 'Connection', id: str, error_msg: str | None = None
) -> Response:
    """Get available and current engine settings for a project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/projects/{id}/settings/engine',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting project startup settings.")
def get_projects_on_startup(
    connection: 'Connection',
    error_msg: str | None = None,
    whitelist: list | None = None,
) -> Response:
    """Get a list of projects along with the nodes on which they would be
    available when the iServer starts up.

    Args:
        connection (Connection): Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list, optional): list of tuples of I-Server Error and
            HTTP errors codes respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/projects/settings/onStartup',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error updating project startup settings.")
def update_projects_on_startup(
    connection: 'Connection',
    body: dict,
    error_msg: str | None = None,
    whitelist: list | None = None,
) -> Response:
    """Update status of projects on iServer nodes at start up. You provide
    the request body as of list of replace operations to be performed on the
    value of array of nodes with the path URI containing the corresponding
    project id that needs to be updated.

    Args:
        connection (Connection): Strategy One REST API connection object
        body (dict): JSON-formatted data used to update project startup settings
            {
            "operationList": [
                {
                    "op": "replace",
                    "path": "/projects/B7CA92F04B9FAE8D941C3E9B7E0CD754/nodes",
                    "value": ["env-183260laio2use1"],
                }
            ]
            }
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list, optional): list of tuples of I-Server Error and
            HTTP errors codes respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(
        endpoint='/api/projects/settings/onStartup',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error getting VLDB settings for project with ID {id}")
def get_vldb_settings(
    connection: 'Connection', id: str, error_msg: str | None = None
) -> Response:
    """Get advanced VLDB settings for a project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/model/projects/{id}?showAdvancedProperties=true',
        headers={'X-MSTR-ProjectID': id},
    )


@ErrorHandler(err_msg="Error updating VLDB settings for project with ID {id}")
def update_vldb_settings(
    connection: 'Connection', id: str, body: dict, error_msg: str | None = None
) -> Response:
    """Update metadata of advanced VLDB settings for a project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        body (dict): JSON-formatted data used to update VLDB settings
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.put(
            endpoint=f'/api/model/projects/{id}',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=body,
        )


@ErrorHandler(
    err_msg="Error getting metadata of VLDB settings for project with ID {id}"
)
def get_applicable_vldb_settings(
    connection: 'Connection', id: str, error_msg: str | None = None
) -> Response:
    """Get metadata of advanced VLDB settings for a project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/model/projects/{id}/applicableAdvancedProperties'
    )


@ErrorHandler(err_msg="Error deleting project with ID {id}")
def delete_project(
    connection: 'Connection', id: str, error_msg: str | None = None
) -> Response:
    """Delete a project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.delete(endpoint=f'/api/projects/{id}')


@ErrorHandler(err_msg="Error getting languages for project with ID {id}")
def get_project_languages(
    connection: 'Connection',
    id: str,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get languages for a specified project

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (string): Project ID
        fields (string, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/projects/{id}/languages',
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error updating languages for project with ID {id}")
def update_project_languages(
    connection: 'Connection',
    id: str,
    body: dict,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Update project language configurations.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        body (dict): JSON-formatted data used to update Project Languages
        fields (string, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(
        endpoint=f'/api/projects/{id}/languages',
        json=body,
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error getting project lock status for project with ID {id}")
def get_project_lock(
    connection: 'Connection',
    id: str,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get the lock status of a project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        fields (string, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/projects/{id}/lock', params={'fields': fields}
    )


@ErrorHandler(err_msg="Error updating project lock status for project with ID {id}")
def update_project_lock(
    connection: 'Connection', id: str, body: dict, error_msg: str | None = None
) -> Response:
    """Update the lock status of a project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        body (dict): JSON-formatted data used to update Project Lock
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(endpoint=f'/api/projects/{id}/lock', json=body)


@ErrorHandler(err_msg="Error unlocking project lock for project with ID {id}")
def delete_project_lock(
    connection: 'Connection',
    id: str,
    lock_type: str,
    force: bool = False,
    lock_id: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Unlock a project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        lock_type (str): The type of the lock to unlock
        force (bool, optional): Whether to force unlock the lock
        lock_id (str, optional): The id of the lock to unlock when
            not force unlock
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.delete(
        endpoint=f'/api/projects/{id}/lock',
        params={'lockType': lock_type, 'force': force, 'lockId': lock_id},
    )


@ErrorHandler(err_msg="Error updating project with ID {id}")
def update_project(
    connection: 'Connection',
    id: str,
    body: dict,
    fields: str | None = None,
) -> 'Response':
    """Update a project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        body (dict): JSON-formatted data used to update Project
        fields (string, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(
        endpoint=f'/api/projects/{id}',
        json=body,
        headers={
            'Content-Type': 'application/json-patch+json',
        },
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error getting project duplications")
def get_project_duplications(
    connection: 'Connection',
    offset: int = 0,
    limit: int = 100,
    error_msg: str | None = None,
) -> Response:
    """Get project duplications.

    Args:
        connection (Connection): Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. 200 on success.
    """
    return connection.get(endpoint='/api/projectDuplications')


@ErrorHandler(err_msg="Error triggering a new project duplication")
def trigger_project_duplication(
    connection: 'Connection',
    id: str,
    body: dict,
    prefer: str = 'respond-async',
    error_msg: str | None = None,
) -> Response:
    """Trigger a new project duplication.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        body (dict): Dictionary containing the project duplication settings
        prefer (str, optional): The preferred response mode. Defaults to
            'respond-async'
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. 201 on success.
    """
    return connection.post(
        endpoint='/api/projectDuplications',
        json=body,
        headers={'X-MSTR-ProjectID': id, 'Prefer': prefer},
    )


@ErrorHandler(err_msg="Error getting project duplication with ID {id}")
def get_project_duplication(
    connection: 'Connection',
    id: str,
    error_msg: str | None = None,
) -> Response:
    """Get a specific project duplication.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project duplication ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. 200 on success.
    """
    return connection.get(endpoint=f'/api/projectDuplications/{id}')


@ErrorHandler(err_msg="Error cancelling project duplication with ID: {id}")
def cancel_project_duplication(
    connection: 'Connection',
    id: str,
    body: dict,
    error_msg: str | None = None,
) -> Response:
    """Cancel a specific project duplication.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project duplication ID
        body (dict): Dictionary containing status update properties
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. 204 on success.
    """
    return connection.put(endpoint=f'/api/projectDuplications/{id}/status', json=body)



from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg='Error sending an email')
def send_email(connection: 'Connection', body: dict, error_msg: str | None = None):
    """Send an email to specified recipients.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (json): JSON-formatted data used to send an email
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.post(endpoint='/api/emails', json=body)



__title__ = "mstrio-py"
__version__ = "11.5.10.101"  # NOSONAR
__license__ = "DUAL-TIER LICENSE AGREEMENT"
__description__ = "Python interface for the Strategy One REST API"
__author__ = "Strategy One"
__author_email__ = "bkaczynski@strategy.com"

from .utils.dev_helpers import what_can_i_do_with  # noqa: F401
__build__ = "11.5.1000.00033"



from requests import Response
from requests.exceptions import SSLError

from mstrio.utils.helper import response_handler


def server_status(connection) -> 'Response | None':
    """
     Args:
        connection: Strategy One REST API connection object
    Returns:
        Complete HTTP response object
    """

    try:
        response = connection.get(skip_expiration_check=True, endpoint='/api/status')
    except SSLError as exc:
        raise SSLError(
            "SSL certificate error.\nPlease double check that the link you "
            "are using comes from a trusted source. If you trust the URL "
            "provided please specify parameter 'ssl_verify' to 'False' in the "
            "'Connection' class.\n\nCheck readme for more details."
        ) from exc

    if not response.ok and response.status_code != 401:
        response_handler(response, "Failed to check server status")
    elif response.status_code == 401:
        return None
    return response
